from dataclasses import dataclass
from math import nan, isnan


@dataclass
class AnswersData:
    checked: set[int]
    correct: set[int]
    all: set[int]

    @property
    def unchecked(self):
        return self.all - self.checked

    @property
    def incorrect(self):
        return self.all - self.correct


@dataclass
class ScoreData:
    correct: float
    skipped: float
    incorrect: float = nan

    def __post_init__(self):
        if isnan(self.incorrect):
            self.incorrect = -self.correct


def some(answers: AnswersData, score: ScoreData) -> float:
    """Return the maximal score if any of the correct answers and no incorrect answer is checked.

    If an incorrect answer is checked, return minimal score."""
    # Answer is valid if and only if :
    # (proposed ≠ ∅ and proposed ⊆ correct) or (proposed = correct = ∅)
    ok = (answers.checked and answers.checked.issubset(answers.correct)) or (
        not answers.checked and not answers.correct
    )
    if ok:
        return score.correct
    elif not answers.checked:
        return score.skipped
    else:
        return score.incorrect


def all(answers: AnswersData, score: ScoreData) -> float:
    """Return the maximal score if all the correct answers are checked, else give the minimal score."""
    ok = answers.checked == answers.correct
    if ok:
        return score.correct
    elif not answers.checked:
        return score.skipped
    else:
        return score.incorrect


def proportional(answers: AnswersData, score: ScoreData) -> float:
    """Returned score is a pondered mean of minimal and maximal scores.

    Minimal and maximal scores are defined respectively in the header of the ptyx file
    with `correct` and `incorrect` keywords.

    The score returned is (1 - k) * minimal_score + k * maximal_score,
    where k is the proportion of correct answers (i.e. boxes left blank when proposed answer
    was incorrect and boxes checked when proposed answer was correct).
    """
    count_ok = len(answers.checked & answers.correct) + len(answers.unchecked & answers.incorrect)
    proportion = count_ok / len(answers.all)
    return round(score.incorrect + proportion * (score.correct - score.incorrect), 2)


# def floored_proportional(answers: AnswersData, score: ScoreData) -> float:
#     return max(0, proportional(answers, score))


# def floored_half_proportional(answers: AnswersData, score: ScoreData) -> float:
#     if (answers.checked == answers.correct):
#         return score.correct
#     else:
#         return 0.5*floored_proportional(answers, score)


def _checked_among_correct_proportion(answers: AnswersData) -> float:
    """Return the proportion of checked answers among correct ones.

    If there was no correct answer, return 1.0 if no answer was checked, 0.0 else.
    """
    if len(answers.correct) == 0:
        return float(len(answers.checked) == 0)
    return len(answers.checked & answers.correct) / len(answers.correct)


def partial_answers_linear(answers: AnswersData, score: ScoreData) -> float:
    """Return 0 if an incorrect answer was checked, else the proportion of correct answers"""
    if answers.checked & answers.incorrect:
        return score.incorrect
    else:
        return round(_checked_among_correct_proportion(answers) * score.correct, 2)


def partial_answers_quadratic(answers: AnswersData, score: ScoreData) -> float:
    if answers.checked & answers.incorrect:
        return score.incorrect
    else:
        return round(_checked_among_correct_proportion(answers) ** 2**score.correct, 2)


# def floored_partial_anwers(answers: AnswersData, score: ScoreData) -> float:
#     return max(0, partial_answers(answers, score))


modes = (some, all, proportional, partial_answers_linear, partial_answers_quadratic)



r"""
ptyx MCQ

This extension enables computer corrected quizzes.

An example:

    #LOAD{mcq}
    #SEED{8737545887}

    ===========================
    sty=my_custom_sty_file
    scores=1 0 0
    mode=all
    ids=~/my_students.csv
    ===========================


    <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

    ======= Mathematics ===========

    * 1+1 =
    - 1
    + 2
    - 3
    - 4

    - another answer

    ======= Litterature ==========

    * "to be or not to be", but who actually wrote that ?
    + W. Shakespeare
    - I. Newton
    - W. Churchill
    - Queen Victoria
    - Some bloody idiot

    * Jean de la Fontaine was a famous French
    - pop singer
    - dancer
    + writer
    - detective
    - cheese maker

    > his son is also famous for
    @{\color{blue}%s}
    - dancing french cancan
    - conquering Honolulu
    - walking for the first time on the moon
    - having breakfast at Tiffany

    + none of the above is correct

    >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


One may include some PTYX code of course.

    """

import re
from pathlib import Path
from re import Match

from ptyx.extensions import extended_python
from ptyx.latex_generator import Compiler

from .compile.extend_latex_generator import MCQLatexGenerator
from .compile.generate_ptyx_code import generate_ptyx_code


# Note for closing tags:
# '@END' means closing tag #END must be consumed, unlike 'END'.
# So, use '@END_QUESTIONS_BLOCK' to close QUESTIONS_BLOCK,
# but use 'END_QUESTIONS_BLOCK' to close QUESTION, since
# #END_QUESTIONS_BLOCK must not be consumed then (it must close
# QUESTIONS_BLOCK too).


__tags__ = {
    # Tags used to structure MCQ
    "QCM": (0, 0, ["@END_QCM"]),
    "SECTION": (0, 0, ["SECTION", "END_QCM"]),
    "NEW_QUESTION": (0, 0, ["NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "CONSECUTIVE_QUESTION": (0, 0, ["NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "VERSION": (1, 0, ["VERSION", "NEW_QUESTION", "CONSECUTIVE_QUESTION", "SECTION", "END_QCM"]),
    "ANSWERS_BLOCK": (0, 0, ["@END_ANSWERS_BLOCK"]),
    "NEW_ANSWER": (2, 0, ["NEW_ANSWER", "END_ANSWERS_BLOCK"]),
    "ANSWERS_LIST": (2, 0, None),
    # Other tags
    "QCM_HEADER": (1, 0, None),
    "DEBUG_MCQ": (0, 0, None),
    # Deprecated tags
    "L_ANSWERS": (1, 0, None),
}
__latex_generator_extension__ = MCQLatexGenerator


def main(text: str, compiler: Compiler) -> str:
    # Generation algorithm is the following:
    # 1. Parse AutoQCM code, to convert it to plain pTyX code.
    #    Doing this, we now know the number of questions, the number
    #    of answers per question and the students names.
    #    However, we can't know for know the number of the correct answer for
    #    each question, since questions numbers and answers numbers too will
    #    change during shuffling, when compiling pTyX code (and keeping track of
    #    them through shuffling is not so easy).
    # 2. Generate syntax tree, and then compile pTyX code many times to generate
    #    one test for each student. For each compilation, keep track of correct
    #    answers.
    #    All those data are stored in `latex_generator.mcq_data['answers']`.
    #    `latex_generator.mcq_data['answers']` is a dict
    #    with the following structure:
    #    {1:  [          <-- test n°1 (test id is stored in NUM)
    #         [0,3,5],   <-- 1st question: list of correct answers
    #         [2],       <-- 2nd question: list of correct answers
    #         [1,5],     ...
    #         ],
    #     2:  [          <-- test n°2
    #         [2,3,4],   <-- 1st question: list of correct answers
    #         [0],       <-- 2nd question: list of correct answers
    #         [1,2],     ...
    #         ],
    #    }

    remove_comments = compiler.syntax_tree_generator.remove_comments

    # First pass, only to include files.
    def include(match: Match) -> str:
        file_found = False
        pattern = match.group(1).strip()
        contents = []
        path: Path
        for path in sorted(compiler.dir_path.glob(pattern)):
            if path.is_file():
                file_found = True
                with open(path) as file:
                    file_content = remove_comments(file.read().strip())
                    if file_content[:2].strip() != "*":
                        file_content = "*\n" + file_content
                    lines = []
                    for line in file_content.split("\n"):
                        lines.append(line)
                        if (
                            line.startswith("* ")
                            or line.startswith("> ")
                            or line.startswith("OR ")
                            or line.rstrip() in ("*", ">", "OR")
                        ):
                            prettified_path = path.parent / f"\u001b[36m{path.name}\u001b[0m"
                            lines.append(f'#PRINT{{\u001b[36mIMPORTING\u001b[0m "{prettified_path}"}}')
                    contents.append("\n".join(lines))
        if not file_found:
            print(f"WARNING: no file corresponding to {pattern!r} !")
        return "\n\n" + "\n\n".join(contents) + "\n\n"

    text = re.sub(r"^-- (.+)$", include, text, flags=re.MULTILINE)

    # Call extended_python extension.
    text = extended_python.main(text, compiler)

    code = generate_ptyx_code(text)
    assert isinstance(code, str)
    return code



# from enum import Enum, unique
#
# @unique
# class Levels(Enum):
#     ROOT, QCM, SECTION, QUESTION, VERSION, ANSWERS_BLOCK, NEW_ANSWER = range(7)


def generate_ptyx_code(text: str) -> str:
    """This function translates MCQ syntax into proper pTyX code."""

    # TODO: improve ability to customize this part ?

    code = []

    level_names = ("ROOT", "QCM", "SECTION", "QUESTION", "VERSION", "ANSWERS_BLOCK", "NEW_ANSWER")
    # noinspection PyTypeChecker
    levels = dict(enumerate(level_names))
    depth = {name: i for i, name in enumerate(level_names)}
    #    stack = StaticStack(levels)
    current_level = levels[0]

    def next_level(level):
        return levels[depth[level] + 1]

    def previous_level(level):
        return levels[depth[level] - 1]

    def begin(level, **kw):
        nonlocal current_level

        if level not in level_names:
            raise RuntimeError(f"Unknown level: {level}")

        # Current level must be the parent one before opening level.
        target = depth[level] - 1
        while depth[current_level] > target:
            close(current_level)
        while depth[current_level] < target:
            begin(next_level(current_level))
        assert depth[current_level] == target

        current_level = level
        content = []
        if level == "QUESTION":
            content.append("#CONSECUTIVE_QUESTION" if kw.get("consecutive") else "#NEW_QUESTION")
        else:
            content.append(f"#{level}")

        if level == "QCM":
            if kw.get("shuffle", True):
                # Shuffle sections.
                content.append("[shuffle]")
        elif level == "SECTION":
            if "title" in kw:
                content.append("[%s]" % kw["title"])
        elif level == "VERSION":
            content.append("{%s}" % kw["n"])
        elif level == "NEW_ANSWER":
            content.append("{%s}{%s}" % (kw["n"], kw["correct"]))

        code.append("".join(content))

    def close(level):
        """Close `level` (and any opened upper one).

        For example, close('SECTION') will close levels until returning
        to a level lower than SECTION ('ROOT' or 'QCM').
        Any opened upper level ('QUESTION_BLOCK' or 'ANSWERS') will be closed first.
        """
        nonlocal current_level

        if depth[current_level] < depth[level]:
            raise RuntimeError(f"I can not close level {level}, since it is not opened !")
        while depth[current_level] > depth[level]:
            close(current_level)

        current_level = previous_level(level)
        if level == "QCM":
            code.append("#END_QCM")

        elif level == "ANSWERS_BLOCK":
            # If there are any blank lines after an answer, they must appear
            # *after* #END_ANSWERS_BLOCK, so move them.
            i = 0
            while code[-1].strip() == "":
                code.pop()
                i += 1
            code.append("#END_ANSWERS_BLOCK")
            code.extend(i * [""])
            # Note that since a single blank line is used to separate answers
            # blocks, user must use two consecutive blanks lines to generate
            # a new LaTeX paragraph (the first one is automatically stripped).
            # XXX: This must appear in doc.

    previous_line = None
    before_QCM = True
    is_header = False
    header = ["#QCM_HEADER{"]
    question_num = 0

    # Don't use ASK_ONLY: if one insert Python code here, it would be removed
    # silently when generating the pdf files with the answers !
    intro = ["#ASK % (introduction)"]

    answer_num = None
    for _line_ in text.split("\n"):
        line = _line_.strip()
        n = len(line)

        if n >= 3 and all(c == "<" for c in line):  # <<<
            # start MCQ
            header.append("}")
            code.extend(header)

            intro.append("#END % (introduction)")
            code.extend(intro)
            print("Parsing QCM...\n")
            print("STRUCTURE:\n")
            begin("QCM")
            before_QCM = False

        elif before_QCM:
            if n >= 3 and all(c == "=" for c in line):  # ===
                # Enter (or leave) header section.
                is_header = not is_header
            elif is_header:
                header.append(_line_)
            else:
                intro.append(_line_)

        elif n >= 3 and line.startswith("=") and line.endswith("="):
            # === title ===
            # Start a new section.
            begin("SECTION", title=line.strip("= "))

        # Nota: for a new version of a question, line must start with 'OR ',
        # with a trailing space, or line must be 'OR', without trailing space.
        elif line[:2].strip() in ("*", ">") or line[:3].strip() == "OR":
            # * question
            # Start a question block, with possibly several versions of a question.

            if line[:2] != "OR":
                # If line starts with 'OR', this is not a new block, only another
                # version of current question block.
                # In all other cases, this is a new question.

                begin("QUESTION", consecutive=(line[0] == ">"))
            question_num += 1
            begin("VERSION", n=question_num)
            answer_num = 0
            code.append(line[2:])

        elif line.startswith("#ANSWERS_LIST"):
            # End question.
            # (Usually, questions are closed when seeing answers, i.e. lines
            # introduced by '-' or '+').
            code.append(line)

        elif line.startswith("@"):
            raw = line.startswith("@@")
            formatting = line[(2 if raw else 1) :].strip()
            if formatting == "":
                formatting = "%s"
            # Declare function to be applied to all answers.
            code.append(f"#{{RAW_CODE={raw};APPLY_TO_ANSWERS={formatting!r};}}")

        elif line.startswith("- ") or line.startswith("+ ") or line.startswith("! "):
            # - incorrect answer
            # + correct answer
            # ! neutralized answer (neither really true nor false, this is useful when there was a problem
            # in an answer).

            if previous_line is None:
                raise RuntimeError("No question before answers list !")

            # A blank line is used to separate answers groups.
            if previous_line == "" and current_level == "NEW_ANSWER":
                # This blank line should not appear in final pdf, so remove it.
                # (NB: This must *not* be done for the first answer !)
                code.pop()

            if previous_line == "" or previous_line.startswith("@"):
                # Answers are shuffled inside their respective groups,
                # however groups are kept separate.
                if previous_line.startswith("@"):
                    cut_and_paste = code.pop()
                begin("ANSWERS_BLOCK")
                if previous_line.startswith("@"):
                    # noinspection PyUnboundLocalVariable
                    code.append(cut_and_paste)

            if answer_num is None:
                raise RuntimeError("No question before answers list !")
            answer_num += 1
            match line[0]:
                case "+":
                    correct = True
                case "-":
                    correct = False
                case "!":
                    correct = None
                case _:
                    assert False, f"{line[0]} should be either '+', '-' or '!'."
            begin("NEW_ANSWER", n=answer_num, correct=correct)

            code.append(line[2:])

        elif n >= 3 and all(c == ">" for c in line):  # >>>
            # End MCQ
            close("QCM")

        else:
            code.append(_line_)

        previous_line = line

    code.append(r"\cleardoublepage")
    code.append(r"\end{document}")
    return "\n".join(code)



from typing import Literal, Iterator, Iterable, Tuple

from numpy import array, nonzero, transpose, ndarray

from ptyx_mcq.scan.visual_debugging import color2debug


Pixel = Tuple[int, int]


# def top_left_iterator(stop, step=1):
#     """Return an iterator for coordinates starting from top-left corner."""
#     # Pixels are visited starting from top-left corner
#     # in the following order:
#     # 1  3  8  15
#     # 4  2  6  13
#     # 9  7  5  11
#     # 16 14 12 10
#     for n in range(0, stop, step):
#         yield n, n
#         for k in range(n - step, -1, -step):
#             yield k, n
#             yield n, k


# def total_grayness(m):
#     return numpy.interp(m, [0, 0.2, 0.8, 1], [0, 0.1, 0.9, 1]).sum()


def find_black_rectangle(
    matrix: ndarray,
    width: int = 50,
    height: int = 50,
    error: float = 0.30,
    gray_level: float = 0.4,
    mode: Literal["row", "column"] = "row",
    debug=False,
) -> Iterator[Pixel]:
    """Detect a black rectangle of given size (in pixels) in matrix.

    The n*m matrix must contain only floats between 0 (white) and 1 (black).

    Optional parameters:
        - `error` is the ratio of white pixels allowed in the black square.
        - `gray_level` is the level above which a pixel is considered to be white.
           If it is set to 0, only black pixels will be considered black ; if it
           is close to 1 (max value), almost all pixels are considered black
           except white ones (for which value is 1.).
        - `mode` is either:
            * 'row' (picture is scanned row by row, from top to bottom)
            * 'column' (picture is scanned column by column, from left to right)

    Return a generator of (i, j) where i is line number and j is column number,
    indicating black squares top left corner.

    """
    # ~ color2debug(matrix)
    # First, convert grayscale image to black and white.
    if debug:
        print(
            f"`find_black_rectangle` parameters: width={width}, "
            f"height={height}, error={error}, gray_level={gray_level}"
        )
        color2debug(matrix, display=True)
    m = array(matrix, copy=False) < gray_level
    if debug:
        color2debug(1 - m, display=True)
    # Black pixels are represented by False, white ones by True.
    # pic_height, pic_width = m.shape
    per_line = (1 - error) * width
    per_col = (1 - error) * height
    goal = per_line * height
    # List of areas to avoid.
    to_avoid: list[tuple[int, int, int, int]] = []
    # Find a black pixel, starting from top left corner,
    # and scanning line by line (i.e. from top to bottom).
    if mode == "row":
        black_pixels: Iterable = nonzero(m)
    elif mode == "column":
        black_pixels = reversed(nonzero(transpose(array(m))))
    else:
        raise RuntimeError("Unknown mode: %s. Mode should be either 'row' or 'column'." % repr(mode))
    if debug:
        print(mode, black_pixels)
    for i, j in zip(*black_pixels):
        # Avoid detecting an already found square.
        if debug:
            print("Black pixel found at %s, %s" % (i, j))
        # ~ color2debug(m.astype(float), (i, j), (i + width, j + height), color=(255,0,255))
        if any(
            (li_min <= i <= li_max and co_min <= j <= co_max) for (li_min, li_max, co_min, co_max) in to_avoid
        ):
            continue
        assert m[i, j] == 1
        total = m[i : i + height, j : j + width].sum()
        if debug:
            print(f"Total: {total} | Goal: {goal}")
            if total >= goal / 5:
                color2debug(matrix, (i, j), (i + height, j + width), display=True)
        # ~ print(f'Black pixels ratio: {total}/{width*height} ; Min: {goal}.')
        # ~ input('- pause -')
        #        print("Detection: %s found (minimum was %s)." % (total, goal))
        if total >= goal:
            # ~ print("\nBlack square found at (%s,%s)." % (i, j))
            # Adjust detection if top left corner is a bit "damaged"
            # (i.e. if some pixels are missing there), or if this pixel is
            # only an artefact before the square.
            if debug:
                color2debug(matrix, (i, j), (i + 2, j + 2), fill=True)
            i0 = i
            j0 = j
            # Note: limit adjustment range (in case there are two consecutive squares)
            for _i in range(50):
                horizontal = vertical = False
                # Horizontal adjustment:
                try:
                    while abs(j - j0) < error * width and (
                        m[i : i + height, j + width + 1].sum() > per_col > m[i : i + height, j].sum()
                    ):
                        j += 1
                        if debug:
                            print("j+=1")
                        horizontal = True
                except IndexError:
                    pass
                # If square was already shifted horizontally in one way, don't try
                # to shift it in the opposite direction.
                if not horizontal:
                    try:
                        while (
                            abs(j - j0) < error * width
                            and m[i : i + height, j + width].sum() < per_col < m[i : i + height, j - 1].sum()
                        ):
                            j -= 1
                            if debug:
                                print("j-=1")
                            horizontal = True
                    except IndexError:
                        pass
                # Vertical adjustment:
                try:
                    while (
                        abs(i - i0) < error * height
                        and m[i + height + 1, j : j + width].sum() > per_line > m[i, j : j + width].sum()
                    ):
                        i += 1
                        if debug:
                            print("i+=1")
                        vertical = True
                except IndexError:
                    pass
                try:
                    while (
                        abs(i - i0) < error * height
                        and m[i + height, j : j + width].sum() < per_line < m[i - 1, j : j + width].sum()
                    ):
                        i -= 1
                        if debug:
                            print("i-=1")
                        vertical = True
                except IndexError:
                    pass
                if not (vertical or horizontal):
                    break
            else:
                print("Warning: adjustment of square position seems abnormally long... Skipping...")
            #
            #      Do not detect pixels there to avoid detecting
            #      the same square twice.
            #      ←—————————————————————————————————→
            #      ←——————————→  ←———————————————————→
            #      buffer zone      square itself
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ///////////   #####################
            #      ←——————————→  ←———————————————————→
            #      ≃ error*size          size

            # Avoid detecting an already found square.
            if any(
                (li_min <= i <= li_max and co_min <= j <= co_max)
                for (li_min, li_max, co_min, co_max) in to_avoid
            ):
                continue

            to_avoid.append((i - error * height - 1, i + height - 2, j - error * width - 1, j + width - 2))
            # ~ print("Final position of this new square is (%s, %s)" % (i, j))
            # ~ print("Forbidden areas are now:")
            # ~ print(to_avoid)
            if debug:
                input("-- pause --")
            yield i, j


def find_black_square(
    matrix: ndarray, size: int, error: float = 0.4, gray_level: float = 0.4, **kw
) -> Iterator[Pixel]:
    return find_black_rectangle(matrix, width=size, height=size, error=error, gray_level=gray_level, **kw)


# def detect_all_squares(matrix, size=50, error=0.30):
#     return list(find_black_square(matrix, size=size, error=error))


def test_square_color(
    m: ndarray,
    i: int,
    j: int,
    size: int,
    proportion: float = 0.3,
    gray_level: float = 0.75,
    margin: int = 0,
    _debug=False,
) -> bool:
    """Return True if square is black, False else.

    (i, j) is top left corner of the square, where i is line number
    and j is column number.
    `proportion` is the minimal proportion of black pixels the square must have
    to be considered black (`gray_level` is the level below which a pixel
    is considered black).
    """
    if size <= 2 * margin + 4:
        raise ValueError("Square too small for current margins !")
    square = m[i + margin : i + size - margin, j + margin : j + size - margin] < gray_level
    if _debug:
        print(square, square.sum(), len(square) ** 2)
        print(
            "proportion of black pixels detected: %s (minimum required was %s)"
            % (square.sum() / size**2, proportion)
        )
    # Test also the core of the square, since borders may induce false
    # positives if proportion is kept low (like default value).
    core = square[2:-2, 2:-2]
    return square.sum() > proportion * len(square) ** 2 and core.sum() > proportion * len(core) ** 2


def eval_square_color(m: ndarray, i: int, j: int, size: int, margin: int = 0, _debug=False) -> float:
    """Return an indicator of blackness, which is a float in range (0, 1).
    The bigger the float returned, the darker the square.

    This indicator is useful to compare several squares, and find the blacker one.
    Note that the core of the square is considered the more important part to assert
    blackness.

    (i, j) is top left corner of the square, where i is line number
    and j is column number.
    """
    if size <= 2 * margin:
        raise ValueError("Square too small for current margins !")
    # Warning: pixels outside the sheet shouldn't be considered black !
    # Since we're doing a sum, 0 should represent white and 1 black,
    # so as if a part of the square is outside the sheet, it is considered
    # white, not black ! This explain the `1 - m[...]` below.
    square = 1 - m[i + margin : i + size - margin, j + margin : j + size - margin]
    return square.sum() / (size - margin) ** 2


def adjust_checkbox(
    m: ndarray, i: int, j: int, size: int, level1: float = 0.5, level2: float = 0.6, delta: int = 5
):
    # return (i, j)
    # Try to adjust top edge of the checkbox
    i0, j0 = i, j
    if m[i : i + size, j : j + 1].sum() < level1 * size:
        for i in range(i0 - delta, i0 + delta + 1):
            if m[i : i + size, j : j + 1].sum() > level2 * size:
                break
        else:
            i = i0
    if m[i : i + 1, j : j + size].sum() < level1 * size:
        for j in range(j0 - delta, j0 + delta + 1):
            if m[i : i + 1, j : j + size].sum() > level2 * size:
                break
        else:
            j = j0
    return i, j





# Unit: cm
CALIBRATION_SQUARE_POSITION = 1
CALIBRATION_SQUARE_SIZE = 0.5
SQUARE_SIZE_IN_CM = 0.25
CELL_SIZE_IN_CM = 0.5
PAPER_FORMATS = {"A4": (21, 29.7), "A5": (14.8, 21), "A3": (29.7, 42)}
PAPER_FORMAT = "A4"
MARGIN_LEFT_IN_CM = 1.5
MARGIN_RIGHT_IN_CM = 1.5
MARGIN_TOP_IN_CM = 2.5
MARGIN_BOTTOM_IN_CM = 2



#!/usr/bin/env python3
import re

# --------------------------------------
#               Compiler
#    Generate all MCQ versions
# --------------------------------------
#    PTYX
#    Python LaTeX preprocessor
#    Copyright (C) 2009-2020  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from functools import partial
from pathlib import Path
from typing import TypedDict, Optional, Set, List, Tuple, Dict

from ptyx.printers import sympy2latex
from ptyx.latex_generator import LatexGenerator
from ptyx.syntax_tree import Node, Tag

from ptyx_mcq.tools.config_parser import Configuration

from .header import (
    packages_and_macros,
    ID_band,
    extract_students_id_and_name_from_csv,
    extract_students_name_from_csv,
    student_id_table,
    students_checkboxes,
    IdentifiantError,
)


class MCQCache(TypedDict):
    header: Optional[str]
    check_id_or_name: Optional[str]
    data: Configuration


def _has_option(node: Node, option: str) -> bool:
    options_list = node.options.split(",") if node.options else []
    return option in [opt.strip() for opt in options_list]


def _analyze_IDS(ids: List[str]) -> Tuple[int, int, List[Tuple[str, ...]]]:
    """Given a list of IDs (str), return:
    - the length of an ID (or raise an error if they don't have the same size),
    - the maximal number of different digits in an ID caracter,
    - a list of sets corresponding to the different digits used for each ID caracter.

    >>> _analyze_IDS(['18', '19', '20', '21'])
    (2, 4, [{'1', '2'}, {'8', '9', '0', '1'}])
    """
    lengths = {len(iD) for iD in ids}
    if len(lengths) != 1:
        print(ids)
        raise IdentifiantError("All students ID must have the same length !")
    ID_length = lengths.pop()
    # On crée la liste de l'ensemble des valeurs possibles pour chaque chiffre.
    digits: List[Set[str]] = [set() for _ in range(ID_length)]
    for iD in ids:
        for i, digit in enumerate(iD):
            digits[i].add(digit)

    max_ndigits = max(len(set_) for set_ in digits)
    # Make digits elements immutable (safer...)
    # Don't use frozenset, as it isn't easily serializable to JSON.
    return ID_length, max_ndigits, [tuple(s) for s in digits]


def _detect_ID_format(ids: Dict[str, str], id_format: str) -> Configuration:
    """Return IDs and ID format data.

    `ids` is a dictionary who contains students names and ids.
    `id_format` is a string, specifying a number of digits ('8 digits'...).

    Returned ID format data will consist of:
    - the length of an ID,
    - the maximal number of different digits in an ID character,
    - a list of sets corresponding to the different digits used for each ID character.
    """
    ID_length = max_ndigits = digits = None

    if not ids and not id_format:
        raise RuntimeError("Unknown format for students' IDs.")
    if ids:
        # Analyze the IDs list even if `id_format` is provided.
        # This enables to check the consistency between the IDs list and the
        # given ID format.
        ID_length, max_ndigits, digits = _analyze_IDS(list(ids))

    if id_format:
        # Analyze the ID format even if students ids are provided.
        # This enables to check the consistency between the IDs list and the
        # given ID format.
        num, ext = id_format.split()
        # Test format syntax
        if ext not in ("digit", "digits"):
            raise ValueError(f"Unknown format : {id_format!r}")
        try:
            n = int(num)
        except ValueError:
            raise ValueError(f"Unknown format : {id_format!r}")
        # Test consistency between format and IDs
        if ID_length is not None and ID_length != n:
            raise IdentifiantError("Identifiers don't match given format !")
        # Generate format data
        ID_length = n
        max_ndigits = 10
        digits = n * [tuple("0123456789")]
    assert ID_length is not None
    assert max_ndigits is not None
    assert digits is not None
    return {"students_ids": ids, "id_format": (ID_length, max_ndigits, digits)}


class SameAnswerError(RuntimeError):
    pass


# Extend LatexGenerator with new tags.


class MCQLatexGenerator(LatexGenerator):
    """Extension of LatexGenerator handling new tags."""

    def _test_singularity_and_append(self, code: str, answers_list: list) -> str:
        code = code.strip()
        if code in answers_list:
            msg = [
                "ERROR: Same answer proposed twice in MCQ !",
                f"Answer {code!r} appeared at least twice for the same question.",
                "Question was:",
                repr(self.current_question),
                "",
                "Nota: if this is really desired behaviour, insert",
                "following lines in the header of the ptyx file:",
                "#PYTHON",
                "ALLOW_SAME_ANSWER_TWICE=True",
                "#END",
            ]
            n = max(len(s) for s in msg)
            stars = (n + 4) * "*"
            print(stars)
            for s in msg:
                print("* " + s)
            print(stars)
            raise SameAnswerError(
                "Same answer proposed twice in MCQ (see message above for more information) !"
            )
        else:
            answers_list.append(code)
        return code

    def _mcq_shuffle_and_parse_children(
        self, node: Node, children: list = None, target: Tag = "ITEM"
    ) -> None:
        if self.context.get("MCQ_KEEP_ALL_VERSIONS"):
            self._parse_children(node.children)
        else:
            self._shuffle_and_parse_children(node, children, target=target)

    def _mcq_pick_and_parse_children(self, node: Node, children: list = None, target: Tag = "ITEM") -> None:
        if self.context.get("MCQ_KEEP_ALL_VERSIONS"):
            self._parse_children(node.children)
        else:
            self._pick_and_parse_children(node, children, target=target)

    def _mcq_reset_cache(self) -> None:
        """Reset MCQ plugin cache.

        Some tags use cache, for code which don't change between two successive compilation.
        (Typically, this is used for (most of) the header)."""
        self.cache["mcq"] = {
            "header": None,
            "check_id_or_name": None,
            "data": {
                # Default configuration:
                "mode": {"default": "some"},
                "correct": {"default": 1},
                "incorrect": {"default": 0},
                "skipped": {"default": 0},
                "floor": {"default": None},
                "ceil": {"default": None},
                # 'correct_answers': correct_answers, # {1: [4], 2:[1,5], ...}
                "students": [],
                "id_table_pos": None,
                "students_ids": {},
                "ordering": {},
                # {NUM: {'questions': [2,1,3...],
                #        'answers': {1: [(2, True), (1, False), (3, True)...], ...}}, ...}
                "boxes": {},  # {NUM: {'tag': 'p4, (23.456, 34.667)', ...}, ...}
                "id_format": None,
            },
        }
        assert set(self.cache["mcq"]["data"]).issubset(set(Configuration.__annotations__))

    @property
    def mcq_cache(self) -> MCQCache:
        if "mcq" not in self.cache:
            self._mcq_reset_cache()
        return self.cache["mcq"]

    @property
    def mcq_data(self) -> Configuration:
        return self.mcq_cache["data"]

    def _parse_QCM_tag(self, node: Node) -> None:
        self.write("\n")  # A new line is mandatory here if there is no text before MCQ.
        # ~ self.mcq_correct_answers = []
        self.current_question = ""
        self.mcq_data["ordering"][self.NUM] = {"questions": [], "answers": {}}
        #    self.mcq_data['answers'] = {}
        # ~ self.mcq_data['question_num'] =
        # Global context for all the MCQ.
        self.mcq_context = self.context.copy()
        if _has_option(node, "shuffle"):
            self._mcq_shuffle_and_parse_children(node, target="SECTION")
        else:
            self._parse_children(node.children)

    def _parse_SECTION_tag(self, node: Node) -> None:
        title = node.options.strip() if node.options else ""
        if title:
            self.write(r"\section{%s}" % title)
        children = node.children
        # Nota: \begin{enumerate} must not be written just after the
        # beginning of the section, since there may be some explanations
        # between the section title and the first question.
        # So, we have to insert it just before the first NEW_QUESTION.
        try:
            i = self._child_index(children, "NEW_QUESTION")
        except ValueError:
            self._parse_children(children)
            return
        self._parse_children(children[:i])
        self.write(r"\begin{enumerate}[resume]")
        self._mcq_shuffle_and_parse_children(node, children[i:], target="NEW_QUESTION")
        self.write(r"\end{enumerate}")

    def _parse_NEW_QUESTION_tag(self, node: Node) -> None:
        # Each question must be independent, so reset all local variables.
        self.set_new_context(self.mcq_context)
        self._mcq_pick_and_parse_children(node, children=node.children, target="VERSION")

    def _parse_CONSECUTIVE_QUESTION_tag(self, node: Node) -> None:
        # For a consecutive question, the context (ie. local variables) must not be reset.
        self._mcq_pick_and_parse_children(node, children=node.children, target="VERSION")

    def _parse_VERSION_tag(self, node: Node) -> None:
        """A version of a question. Each question have one or more versions.

        Tag usage: #VERSION{num}
        """
        n = int(node.arg(0))
        self.mcq_question_number = n
        # This list is used to test that the same answer is not proposed twice.
        self.mcq_answers: List[str] = []
        data = self.mcq_data["ordering"][self.NUM]
        data["questions"].append(n)
        data["answers"][n] = []
        self.context["APPLY_TO_ANSWERS"] = None
        self.context["RAW_CODE"] = None
        self.write(r"\pagebreak[3]\item\filbreak")
        self.write(r"\setcounter{answerNumber}{0}")

        # This is used to improve message error when an error occurred.
        # The question itself is stored to make debugging easier (error messages will
        # display current question).
        # Use `self.current_question` to get current question.
        def remember_last_question(code):
            self.current_question = code
            return code

        # So, we have to find where the question itself ends and the answers start.
        try:
            i = self._child_index(node.children, "ANSWERS_BLOCK")
        except ValueError:
            try:
                i = self._child_index(node.children, "ANSWERS_LIST")
            except ValueError:
                raise RuntimeError(
                    "No answers found after a MCQ question !\n"
                    "Question text:\n"
                    "----------------------------------\n"
                    f"{node.as_text(skipped_children=(0,)).strip()}\n"
                    "----------------------------------\n"
                )
        self._parse_children(node.children[1:i], function=remember_last_question)
        # This is the end of the question itself.

        # And then, the answers follow.
        self.write("\n\\nopagebreak[4]\n")
        self.write("\n\\begin{minipage}{\\textwidth}\n\\begin{flushleft}")
        self._parse_children(node.children[i:])
        self.write("\\end{flushleft}\n\\end{minipage}")

    def _parse_ANSWERS_BLOCK_tag(self, node: Node) -> None:
        self._mcq_shuffle_and_parse_children(node, target="NEW_ANSWER")

    def _parse_NEW_ANSWER_tag(self, node: Node) -> None:
        """A new answer.

        Tag usage: #NEW_VERSION{num}{is_answer_correct}
        """

        k = int(node.arg(0))
        is_correct: Optional[bool]
        match arg1 := node.arg(1).strip():
            case "True":
                is_correct = True
            case "False":
                is_correct = False
            case "None":
                is_correct = None
            case _:
                raise RuntimeError(f"Second #NEW_ANSWER argument must be True, False or None, not {arg1!r}.")
        n = self.mcq_question_number

        self._open_answer(n, k, is_correct)
        # Functions to apply. Last one is applied first:
        # if functions = [f, g, h], then s -> f(g(h(s))).
        functions = []

        # TODO(?): functions should be compiled only once for each question block,
        # not for every answer (though it is probably not a bottleneck in
        # code execution).
        apply = self.context.get("APPLY_TO_ANSWERS")
        if apply:
            # Apply template or function to every answer.
            # Support:
            # - string templates. Ex: \texttt{%s}
            # - name of the function to apply. Ex: f.
            # In last case, the function must have been defined or imported
            # before.
            if "%s" in apply:
                functions.append(lambda s: (apply % s))
            else:
                # Search for the function name in context.
                functions.append(self.context[apply])

        if self.context.get("RAW_CODE"):
            # Try to emulate verbatim (which is not allowed inside
            # a macro argument in LaTeX).
            def escape(s):
                # Replace \ first !
                s = s.replace("\\", r"\textbackslash<!ø5P3C14Lø?>")
                s = s.replace("~", r"\textasciitilde<!ø5P3C14Lø?>")
                s = s.replace("^", r"\textasciicircum<!ø5P3C14Lø?>")
                s = s.replace("'", r"\textquotesingle<!ø5P3C14Lø?>")
                for char in "#$%&_{}":
                    s = s.replace(char, rf"\{char}")
                s = s.replace("<!ø5P3C14Lø?>", "{}")
                return rf"\texttt{{{s}}}"

            functions.append(escape)

        if not self.context.get("ALLOW_SAME_ANSWER_TWICE"):
            # This function is used to verify that each answer is unique.
            # This avoids proposing twice the same answer by mistake, which
            # may occur easily when using random values.
            functions.append(partial(self._test_singularity_and_append, answers_list=self.mcq_answers))

        self._parse_children(node.children[2:], function=functions)
        self._close_answer()

    def _open_answer(self, n: int, k: int, is_correct: Optional[bool]) -> None:
        # `n` is question number *before* shuffling
        # `k` is answer number *before* shuffling
        # When the pdf with solutions will be generated, incorrect answers
        # will be preceded by a white square, while correct ones will
        # be preceded by a gray one.
        self.write(r"\ptyxMCQTab{")
        cb_id = f"Q{n}-{k}"
        if self.WITH_ANSWERS and is_correct:
            self.write(r"\checkBox{gray}{%s}" % cb_id)
        elif self.WITH_ANSWERS and is_correct is None:
            self.write(r"\checkBox{orange!20}{%s}" % cb_id)
        else:
            self.write(r"\checkBox{white}{%s}" % cb_id)
        self.write(r"}{")
        data = self.mcq_data["ordering"][self.NUM]
        data["answers"][n].append((k, is_correct))

    def _close_answer(self) -> None:
        # Close 'AutoQCMTab{' written by `_parse_NEW_ANSWER_tag()`.
        self.write(r"}\quad" "\n")

    def _parse_ANSWERS_LIST_tag(self, node: Node) -> None:
        """This tag generates answers from a python list.

        Tag usage: #ANSWERS_LIST{list_of_answers}{list_of_correct_answers}

        Example:
        #ANSWERS_LIST{l}{[l[0]]} or #ANSWERS_LIST{l}{l[0],}
        When using the last syntax, the coma is mandatory.

        Note that if the elements of the lists are not strings, they will be
        converted automatically to math mode latex code (1/2 -> '$\frac{1}{2}$').
        """

        def eval_and_format_arg(arg_num) -> List[str]:
            raw_list = eval(node.arg(arg_num).strip(), self.context)
            if not isinstance(raw_list, (list, tuple)):
                raise RuntimeError(f"In #ANSWERS_LIST, argument {arg_num + 1} must be a list of answers.")
            formatted_list = [(val if isinstance(val, str) else f"${sympy2latex(val)}$") for val in raw_list]
            return formatted_list

        answers = eval_and_format_arg(0)
        correct_answers = eval_and_format_arg(1)

        # Test that arguments seem correct
        # (they must be a list of unique answers, and answers must include the correct ones).
        for ans in correct_answers:
            if ans not in answers:
                raise RuntimeError(
                    f"#ANSWERS_LIST: correct answer {ans!r} is not in proposed answers list {answers!r}!"
                )

        for ans in answers:
            self._test_singularity_and_append(ans, self.mcq_answers)
        # answers = self.mcq_answers

        # Shuffle and generate LaTeX.
        # randfunc.shuffle(answers)
        self.write("\n\n" r"\begin{minipage}{\textwidth}" "\n")
        n = self.mcq_question_number
        for k, ans in enumerate(answers, 1):
            self._open_answer(n, k, ans in correct_answers)
            self.write(ans)
            self._close_answer()
        self.write("\n\n\\end{minipage}")

    def _parse_L_ANSWERS_tag(self, node: Node) -> None:
        raise DeprecationWarning(
            "L_ANSWERS tag is not supported anymore.\n"
            "Use #ANSWERS_LIST{list_of_answers}{list_of_correct_answers} instead of "
            "#L_ANSWERS{list_of_answers}{correct_answer}.\n"
            "Example: #L_ANSWERS{l}{l[0]} -> #ANSWERS_LIST{l}{[l[0]]}"
        )

    def _parse_DEBUG_MCQ_tag(self, node: Node) -> None:
        data = self.mcq_data
        print("---------------------------------------------------------------")
        print("ptyxMCQ data:")
        print(data)
        print("---------------------------------------------------------------")
        # self.write(data)

    @staticmethod
    def _parse_sty_list(sty_packages: str) -> str:
        """Parse LaTeX sty packages list in header."""
        packages: list[str] = []
        for package in sty_packages.split(","):
            package = package.strip()
            # noinspection RegExpRedundantEscape
            m = re.match(r"\s*\[(?P<options>(?:\w|\s)+)\](?P<package>.+)", package)
            if m is None:
                packages.append(f"{{{package}}}")
            else:
                options = m.group("options").strip()
                package = m.group("package").strip()
                packages.append(f"[{options}]{{{package}}}")
        return "\n".join(r"\usepackage" + package for package in packages)

    @staticmethod
    def _extract_config_from_header(text: str) -> tuple[dict[str, str], list[str]]:
        def format_key(key_: str) -> str:
            return key_.strip().replace(" ", "_").lower()

        # {alias: standard key name}
        alias = {
            "name": "names",
            "student": "names",
            "students": "names",
            "id": "students_ids",
            "ids": "students_ids",
            "student_ids": "students_ids",
            "student_id": "students_ids",
            "students_id": "students_ids",
            "package": "sty",
            "packages": "sty",
            "id_formats": "id_format",
            "ids_formats": "id_format",
            "ids_format": "id_format",
        }
        # Read config: a dictionary is generated from the `key = value` entries.
        config: Dict[str, str] = {}
        # After the `key = value` entries, the user may append some raw LaTeX code.
        raw_latex = []
        remaining_is_raw_latex = False
        for line in text.split("\n"):
            if not remaining_is_raw_latex:
                if "=" in line:
                    key, val = line.split("=", maxsplit=1)
                    # Normalize the key.
                    key = format_key(key)
                    key = alias.get(key, key)
                    config[key] = val.strip()
                # A line of --- is used as a delimiter between the `key = value` entries
                # and the optional LaTeX code.
                if line.startswith("---"):
                    remaining_is_raw_latex = True
            else:
                raw_latex.append(line)
        return config, raw_latex

    def _parse_QCM_HEADER_tag(self, node: Node) -> None:
        """Parse HEADER.

        HEADER raw format is the following:
        ===========================
        sty=my_custom_sty_file
        correct=1
        incorrect=0
        skipped=0
        mode=all
        ids=~/my_students_ids_and_names.csv
        names=~/my_students_names.csv
        id_format=8 digits
        ---------------------------
        # Custom lines to include at the end of the LaTeX preamble.
        ===========================
        """
        sty = ""
        raw_latex: list[str] = []
        #    if self.WITH_ANSWERS:
        #        self.context['format_ask'] = (lambda s: '')

        check_id_or_name = self.mcq_cache["check_id_or_name"]
        if check_id_or_name is None:
            code = ""

            config, raw_latex = self._extract_config_from_header(node.arg(0))

            for key in ("mode", "correct", "incorrect", "skipped", "floor", "ceil"):
                if key in config:
                    self.mcq_data[key]["default"] = config.pop(key)  # type: ignore

            if "names" in config:
                # the value must be the path of a CSV file.
                csv = config.pop("names")
                if not self.WITH_ANSWERS:
                    students = extract_students_name_from_csv(Path(csv), self.compiler.file_path)
                    code = students_checkboxes(students)
                    self.mcq_data["students_list"] = students

            if "students_ids" in config or "id_format" in config:
                # config['ids'] must be the path of a CSV file.
                csv = config.pop("students_ids", "")
                id_format = config.pop("id_format", "")

                if not self.WITH_ANSWERS:
                    if csv:
                        ids = extract_students_id_and_name_from_csv(Path(csv), self.compiler.file_path)
                    else:
                        ids = {}

                    try:
                        data = _detect_ID_format(ids, id_format)
                    except IdentifiantError as e:
                        msg = e.args[0]
                        raise IdentifiantError(f"Error in {csv!r} : {msg!r}")

                    self.mcq_data.update(data)  # type: ignore
                    code = student_id_table(*data["id_format"])

            if "sty" in config:
                sty = config.pop("sty")

            # Config should be empty by now !
            for key in config:
                if key == "scores":
                    right, wrong, skipped = config[key].split()
                    print("Please update your ptyx file header:")
                    print(f"Replace `scores = {config[key]}` with:")
                    print(20 * "-")
                    print(f"correct = {right}")
                    print(f"incorrect = {wrong}")
                    print(f"skipped = {skipped}")
                    print(20 * "-")
                raise NameError(f"Unknown key {key!r} in the header of the pTyX file.")

            check_id_or_name = code if not self.WITH_ANSWERS else ""
            self.mcq_cache["check_id_or_name"] = check_id_or_name
            check_id_or_name += r"""
            \vspace{1em}

            \tikz{\draw[dotted] ([xshift=2cm]current page.west) -- (current page.east);}
            """

        header = self.mcq_cache["header"]
        if header is None:
            header1, header2 = packages_and_macros()
            header = "\n".join([header1, self._parse_sty_list(sty), header2, *raw_latex, r"\begin{document}"])
            self.mcq_cache["header"] = header

        # Generate barcode
        # Barcode must NOT be put in the cache, since each document has a
        # unique ID.
        n = self.NUM
        calibration = "MCQ__SCORE_FOR_THIS_STUDENT" not in self.context
        barcode = ID_band(doc_id=n, calibration=calibration)

        self.write("\n".join([header, barcode, check_id_or_name]))



#!/usr/bin/env python3

# -----------------------------------------
#                  Scan
#     Extract info from digitized documents
# -----------------------------------------
#    PTYX
#    Python LaTeX preprocessor
#    Copyright (C) 2009-2020  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import csv
import string
import subprocess
import tempfile
from ast import literal_eval
from hashlib import blake2b
from math import inf
from multiprocessing import Pool
from pathlib import Path
from shutil import rmtree
from time import strftime
from typing import Iterator, TypedDict
from typing import Union, Literal, Optional

from PIL import Image
from numpy import int8, array, ndarray

# File `compilation.py` is in ../.., so we have to "hack" `sys.path` a bit.
# script_path = dirname(abspath(sys._getframe().f_code.co_filename))
# sys.path.insert(0, join(script_path, '../..'))

# from ptyx.compilation import join_files, compile_latex

from . import scores
from .amend import amend_all
from .pdftools import extract_pdf_pictures, PIC_EXTS, number_of_pages
from .scan_pic import (
    scan_picture,
    ANSI_YELLOW,
    ANSI_RESET,
    ANSI_CYAN,
    ANSI_GREEN,
    ANSI_RED,
    CalibrationError,
    PicData,
)
from .visual_debugging import color2debug
from .tools import search_by_extension, print_framed_msg
# from ..compile.header import answers_and_score
from ..tools.config_parser import load, get_answers_with_status, Configuration


class MissingQuestion(RuntimeError):
    """Error raised when some questions where not seen when scanning all data."""


class MissingConfigurationData(RuntimeError):
    """Error raised when some configuration data is missing."""


class FilesPaths(TypedDict, total=False):
    base: str
    skipped: Path
    verified: Path
    results: Path
    cfg: Path


class DocumentData(TypedDict):
    pages: dict[int, PicData]
    name: str
    student_ID: str
    answered: dict[int, set[int]]  # {question: set of answers}
    score: float
    score_per_question: dict[int, float]  # {question: score}


def pic_names_iterator(data: dict[int, DocumentData]) -> Iterator[Path]:
    """Iterate over all pics found in data (i.e. all the pictures already analysed)."""
    for doc_data in data.values():
        for pic_data in doc_data["pages"].values():
            path = Path(pic_data["pic_path"])
            # return pdfhash/picnumber.png
            yield path.relative_to(path.parent.parent)


class MCQPictureParser:
    """Main class for parsing pdf files containing all the scanned MCQ."""

    def __init__(
        self,
        path: Union[str, Path],
        input_dir: Optional[Path] = None,
        output_dir: Optional[Path] = None,
    ):
        self.path = Path(path).expanduser().resolve()
        # Main paths.
        self.dirs: dict[str, Path] = {}
        self.files: FilesPaths = {}
        # All data extracted from pdf files.
        self.data: dict[int, DocumentData] = {}
        # Additional information entered manually.
        self.more_infos: dict[int, tuple[str, str]] = {}  # sheet_id: (name, student_id)
        self.config: Configuration = {}
        # Manually verified pages.
        self.verified: set[Path] = set()
        # `name2docID` is used to retrieve the data associated with a name.
        # FORMAT: {name: test ID}
        self.name2docID: dict[str, int] = {}
        # Set `already_seen` will contain all seen (ID, page) couples.
        # It is used to catch a hypothetical scanning problem:
        # we have to be sure that the same page on the same test is not seen
        # twice.
        self.already_seen: set[tuple[int, int]] = set()
        self.skipped: set[Path] = set()
        self.warnings = False
        self.logname = strftime("%Y.%m.%d-%H.%M.%S") + ".log"
        self.correct_answers: dict[int, dict[int, set[int]]] = {}  # {doc_id: {question: set of answers}}
        self._generate_paths(input_dir, output_dir)
        self._load_configuration()

    def _load_data(self) -> None:
        if self.dirs["data"].is_dir():
            for filename in self.dirs["data"].glob("*.scandata"):
                print(f"Loading: {filename}")
                doc_id = int(filename.stem)
                try:
                    with open(filename) as f:
                        self.data[doc_id] = literal_eval(f.read())
                except Exception:
                    print(f"ERROR when reading {filename} :")
                    raise

    def _store_data(self, pdf_hash: str, doc_id: int, p: int, matrix: ndarray = None) -> None:
        with open(self.dirs["data"] / f"{pdf_hash}.index", "a") as f:
            f.write(str(doc_id) + "\n")
        with open(self.dirs["data"] / f"{doc_id}.scandata", "w") as f:
            f.write(repr(self.data[doc_id]))
        # We will store a compressed version of the matrix.
        # (It would consume too much memory else).
        if matrix is not None:
            webp = self.dirs["data"] / f"{doc_id}-{p}.webp"
            Image.fromarray((255 * matrix).astype(int8)).save(str(webp), format="WEBP")

    def get_pic(self, doc_id: int, page: int) -> Image.Image:
        webp = next(self.dirs["data"].glob(f"{doc_id}-{page}.webp"))
        return Image.open(str(webp))

    def get_matrix(self, doc_id: int, page: int) -> ndarray:
        # noinspection PyTypeChecker
        return array(self.get_pic(doc_id, page).convert("L")) / 255

    def _read_name_manually(
        self, doc_id: int, matrix: ndarray = None, p: int = None, default: str = None
    ) -> tuple[str, str]:
        if matrix is None:
            assert p is not None
            matrix = self.get_matrix(doc_id, p)
        student_ids = self.config["students_ids"]
        student_ID = ""
        print("Name can not be read automatically.")
        print("Please read the name on the picture which will be displayed now.")
        input("-- Press enter --")
        #    subprocess.run(["display", "-resize", "1920x1080", pic_path])
        # TODO: use first letters of students name to find student.
        # (ask again if not found, or if several names match first letters)
        process = None
        while True:
            nlines = matrix.shape[1]
            # Don't relaunch process if it is still alive.
            # (process.poll() is not None for dead processes.)
            if process is None or process.poll() is not None:
                process = color2debug(matrix[0 : int(3 / 4 * nlines), :], wait=False)
            name = input("Student name or ID:").strip()
            if not name:
                if default is None:
                    continue
                name = default
            if student_ids:
                if name in student_ids:
                    name, student_ID = student_ids[name], name
                elif any((digit in name) for digit in string.digits):
                    # This is not a student name !
                    print("Unknown ID.")
                    continue
            print("Name: %s" % name)
            if input("Is it correct ? (Y/n)").lower() not in ("y", "yes", ""):
                continue
            if name:
                break
        process.terminate()
        # Keep track of manually entered information (will be useful
        # if `scan.py` has to be run again later !)
        #        self.more_infos[ID] = (name, student_ID)
        with open(self.files["cfg"], "a", newline="") as csvfile:
            writerow = csv.writer(csvfile).writerow
            writerow([str(doc_id), name, student_ID])
        return name, student_ID

    def _test_integrity(self) -> None:
        """For every test:
        - all pages must have been scanned,
        - all questions must have been seen."""
        questions_not_seen = {}
        pages_not_seen = {}
        ordering = self.config["ordering"]
        for doc_id in self.data:
            try:
                doc_ordering = ordering[doc_id]
            except KeyError:
                raise MissingConfigurationData(
                    f"No configuration data found for document #{doc_id}.\n"
                    "Maybe you recompiled the ptyx file in the while ?\n"
                    f"(Launching `mcq make -n {max(self.data)}` might fix it.)"
                )
            questions = set(doc_ordering["questions"])
            diff = questions - set(self.data[doc_id]["answered"])
            if diff:
                questions_not_seen[doc_id] = ", ".join(str(q) for q in diff)
            # All tests may not have the same number of pages, since
            # page breaking will occur at a different place for each test.
            pages = set(self.config["boxes"][doc_id])
            diff = pages - set(self.data[doc_id]["pages"])
            if diff:
                pages_not_seen[doc_id] = ", ".join(str(p) for p in diff)
        if pages_not_seen:
            self._warn("= WARNING =")
            self._warn("Pages not seen:")
            for doc_id in sorted(pages_not_seen):
                self._warn(f"    • Test {doc_id}: page(s) {pages_not_seen[doc_id]}")
        if questions_not_seen:
            self._warn("=== ERROR ===")
            self._warn("Questions not seen !")
            for doc_id in sorted(questions_not_seen):
                self._warn(f"    • Test {doc_id}: question(s) {questions_not_seen[doc_id]}")

        if questions_not_seen:
            # Don't raise an error for pages not found (only a warning in log)
            # if all questions were found, this was probably empty pages.
            raise MissingQuestion("Questions not seen ! (Look at message above).")

    def _keep_previous_version(self, pic_data: PicData) -> bool:
        """Test if a previous version of the same page exist.

        If so, it probably means the page has been scanned twice, but it could
        also indicate a more serious problem (for example, tests with the same ID
        have been given to different students !).
        As a precaution, we should signal the problem to the user, and ask him
        what he wants to do.
        """
        doc_id = pic_data["ID"]
        p = pic_data["page"]

        # This page has never been seen before, everything is OK.
        if (doc_id, p) not in self.already_seen:
            self.already_seen.add((doc_id, p))
            return False

        # This is problematic: it seems like the same page has been seen twice.
        lastpic_path = pic_data["pic_path"]
        lastpic = Path(lastpic_path).relative_to(self.dirs["pic"])
        firstpic_path = self.data[doc_id]["pages"][p]["pic_path"]
        firstpic = Path(firstpic_path).relative_to(self.dirs["pic"])
        assert isinstance(lastpic_path, str)
        assert isinstance(firstpic_path, str)

        self._warn(f"WARNING: Page {p} of test #{doc_id} seen twice " f'(in "{firstpic}" and "{lastpic}") !')
        action = None
        keys = ("name", "student_ID", "answered")
        if all(pic_data[key] == self.data[doc_id]["pages"][p][key] for key in keys):  # type: ignore
            # Same information found on the two pages, just keep one version.
            action = "f"
            self._warn("Both page have the same information, keeping only first one...")

        # We have a problem: this is a duplicate.
        # In other words, we have 2 different versions of the same page.
        # Ask the user what to do.
        while action not in ("f", "l"):
            print("What must we do ?")
            print("- See pictures (s)")
            print("- Keep only first one (f)")
            print("- Keep only last one (l)")

            action = input("Answer:")
            if action == "s":
                with tempfile.TemporaryDirectory() as tmpdirname:
                    path = Path(tmpdirname) / "test.png"
                    # https://stackoverflow.com/questions/39141694/how-to-display-multiple-images-in-unix-command-line
                    subprocess.run(["convert", firstpic_path, lastpic_path, "-append", str(path)], check=True)
                    subprocess.run(["feh", "-F", str(path)], check=True)
                    input("-- pause --")
        # We must memorize which version should be skipped in case user
        # launch scan another time.
        skipped_pic = firstpic if action == "l" else lastpic
        with open(self.files["skipped"], "a", newline="", encoding="utf8") as file:
            file.write(f"{skipped_pic}\n")

        if action == "l":
            # Remove first picture information.
            del self.data[doc_id]["pages"][p]
            self._store_data(firstpic.parent.name, doc_id, p)

        return action == "f"

    def _extract_name(self, doc_id: int, doc_data: DocumentData, matrix: ndarray, ask: bool = False) -> None:
        # TODO: what is matrix type ?
        pic_data = doc_data["pages"][1]
        # (a) The first page should contain the name
        #     ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾
        # However, if the name was already set (using `more_infos`),
        # don't overwrite it.
        if not doc_data["name"]:
            # Store the name read (except if ask not to do so).
            if not ask:
                doc_data["name"] = pic_data["name"]

        name = doc_data["name"]

        # (b) Update name manually if it was not found
        #     ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾

        if not name:
            name = self._read_name_manually(doc_id, matrix)[0]

        # (c) A name must not appear twice
        #     ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾ ‾
        while name in self.name2docID:
            print(f"Test #{self.name2docID[name]}: {name}")
            print(f"Test #{doc_id}: {name}")
            print_framed_msg(
                f"Error : 2 tests for same student ({name}) !\n"
                "Please modify at least one name (enter nothing to keep a name)."
            )
            # Remove twin name from name2doc_id, and get the corresponding previous test ID.
            ID0 = self.name2docID.pop(name)
            # Ask for a new name.
            name0, student_ID0 = self._read_name_manually(ID0, p=1, default=name)
            # Update all infos.
            self.name2docID[name0] = ID0
            self.data[ID0]["name"] = name0
            self.data[ID0]["student_ID"] = student_ID0
            # Ask for a new name for new test too.
            name = self._read_name_manually(doc_id, matrix, default=name)[0]

        assert name, "Name should not be left empty at this stage !"
        self.name2docID[name] = doc_id
        doc_data["name"] = name

    def _calculate_scores(self) -> None:
        cfg = self.config
        default_mode = cfg["mode"]["default"]
        default_correct = cfg["correct"]["default"]
        default_incorrect = cfg["incorrect"]["default"]
        default_skipped = cfg["skipped"]["default"]
        default_floor = cfg["floor"]["default"]
        default_ceil = cfg["ceil"]["default"]

        for doc_id in self.data:
            correct_ans = self.correct_answers[doc_id]
            neutralized_ans = self.neutralized_answers[doc_id]
            print(f'Test {doc_id} - {self.data[doc_id]["name"]}')
            doc_data = self.data[doc_id]
            for q in sorted(doc_data["answered"]):
                answered = set(doc_data["answered"][q])
                correct_ones = correct_ans[q]
                neutralized_ones = neutralized_ans[q]
                all_answers = {ans_num for ans_num, is_ok in cfg["ordering"][doc_id]["answers"][q]}

                # Neutralized answers must be removed from each set of answers.
                # (Typically, neutralized answers are answers which were detected faulty during or after the
                # examination).
                answered -= neutralized_ones
                correct_ones -= neutralized_ones
                all_answers -= neutralized_ones

                mode = cfg["mode"].get(q, default_mode)

                if mode == "skip":
                    # Used mostly to skip bogus questions.
                    print(f"Question {q} skipped...")
                    continue

                try:
                    func = getattr(scores, mode)
                except AttributeError:
                    raise AttributeError(f"Unknown evaluation mode: {mode!r}.")

                ans_data = scores.AnswersData(checked=answered, correct=correct_ones, all=all_answers)
                scores_data = scores.ScoreData(
                    correct=float(cfg["correct"].get(q, default_correct)),
                    skipped=float(cfg["skipped"].get(q, default_skipped)),
                    incorrect=float(cfg["incorrect"].get(q, default_incorrect)),
                )
                earn = func(ans_data, scores_data)

                floor = cfg["floor"].get(q, default_floor)
                assert floor is None or isinstance(floor, (float, int))
                if floor is not None and earn < floor:
                    earn = floor
                ceil = cfg["ceil"].get(q, default_ceil)
                assert ceil is None or isinstance(ceil, (float, int))
                if ceil is not None and earn > ceil:
                    earn = ceil

                if earn == scores_data.correct:
                    color = ANSI_GREEN
                elif earn == scores_data.incorrect:
                    color = ANSI_RED
                else:
                    color = ANSI_YELLOW
                print(f"-  {color}Rating (Q{q}): {color}{earn:g}{ANSI_RESET}")
                doc_data["score"] += earn
                doc_data["score_per_question"][q] = earn

    def generate_output(self) -> None:
        """Generate CSV files with scores and annotated documents."""
        max_score = self.config["max_score"]
        # Generate CSV file with results.
        # TODO: Add ability to change default score ("ABI" for now).
        scores: dict[str, float | str] = {name: "ABI" for name in self.config["students_ids"].values()}
        results: dict[str, float] = {doc_data["name"]: doc_data["score"] for doc_data in self.data.values()}
        scores.update(results)

        def format_(num: float | str):
            if isinstance(num, float):
                num = round(num, 2)
            return num

        # ~ print(scores)
        scores_path = self.dirs["output"] / "scores.csv"
        print(f"{ANSI_CYAN}SCORES (/{max_score:g}):{ANSI_RESET}")
        with open(scores_path, "w", newline="") as csvfile:
            writerow = csv.writer(csvfile).writerow
            writerow(("Name", "Score"))
            for name in sorted(scores):
                score = scores[name]
                print(f" - {name}: {format_(score)}")
                # TODO: Add ability to change the notation system.
                if isinstance(score, float):
                    score = score / max_score * 20
                writerow([name, format_(score)])
        if results:
            mean = round(sum(results.values()) / len(results), 2)
            print(f"{ANSI_YELLOW}Mean: {mean:g}/{max_score:g}{ANSI_RESET}")
        else:
            print("No score found !")
        print(f'\nResults stored in "{scores_path}"\n')

        # Generate CSV file with ID and pictures names for all students.
        info_path = self.dirs["output"] / "infos.csv"
        info = [
            (
                doc_data["name"],
                doc_data["student_ID"],
                doc_id,
                doc_data["score"],
                [doc_data["pages"][p]["pic_path"] for p in doc_data["pages"]],
            )
            for doc_id, doc_data in self.data.items()
        ]
        print(f"{ANSI_CYAN}SCORES (/{max_score:g}):{ANSI_RESET}")
        with open(info_path, "w", newline="") as csvfile:
            writerow = csv.writer(csvfile).writerow
            writerow(("Name", "Student ID", "Test ID", "Score", "Pictures"))
            for name, student_ID, doc_id, score, paths in sorted(info):
                paths_as_str = ", ".join(str(pth) for pth in paths)
                writerow([name, student_ID, f"#{doc_id}", score, paths_as_str])
        print(f'Infos stored in "{info_path}"\n')
        amend_all(self)
        print(f"\n{ANSI_GREEN}Success ! {ANSI_RESET}:)")

    # def generate_correction(self, display_score: bool = True) -> None:
    #     """Generate pdf files, with the score and the table of correct answers for each test."""
    #     pdf_paths: List[Path] = []
    #     for doc_id, doc_data in self.data.items():
    #         # ~ identifier, answers, name, score, students, ids
    #         name = doc_data["name"]
    #         score = doc_data["score"]
    #         prefix = self.files["base"]
    #         path = self.dirs["pdf"] / f"{prefix}-{doc_id}-corr.score"
    #         pdf_paths.append(path)
    #         print(f"Generating pdf file for student {name} (subject {doc_id}, score {score})...")
    #         latex = answers_and_score(self.config, name, doc_id, (score if display_score else None))
    #         texfile_name = path.with_suffix(".tex")
    #         with open(texfile_name, "w") as texfile:
    #             texfile.write(latex)
    #             texfile.flush()
    #         compile_latex(texfile_name, dest=path, quiet=True)
    #     join_files(self.files["results"], pdf_paths, remove_all=True, compress=True)

    def _load_configuration(self) -> None:
        """Read configuration file, load configuration and calculate maximal score too."""
        configfile = search_by_extension(self.path, ".mcq.config.json")
        cfg: Configuration = load(configfile)
        self.correct_answers = get_answers_with_status(cfg, correct=True)
        self.neutralized_answers = get_answers_with_status(cfg, correct=None)
        default_mode = cfg["mode"]["default"]
        default_correct = cfg["correct"]["default"]

        max_score = 0
        # Take a random student test, and calculate max score for it.
        # Maximal score = (number of questions)x(score when answer is correct)
        for q in next(iter(cfg["ordering"].values()))["questions"]:
            if cfg["mode"].get(q, default_mode) != "skip":
                max_score += int(cfg["correct"].get(q, default_correct))
        cfg["max_score"] = max_score
        self.config = cfg

    def _make_dirs(self, reset: bool = False) -> None:
        """Generate output directory structure.

        If `reset` is True, remove all output directory content.
        """
        if reset and self.dirs["output"].is_dir():
            rmtree(self.dirs["output"])

        for directory in self.dirs.values():
            if not directory.is_dir():
                directory.mkdir()

    def _generate_paths(self, input_dir: Optional[Path] = None, output_dir: Optional[Path] = None) -> None:
        root = self.path
        if not root.is_dir():
            root = root.parent
            if not root.is_dir():
                raise FileNotFoundError("%s does not seem to be a directory !" % root)
        self.dirs["root"] = root

        # ------------------
        # Generate the paths
        # ------------------

        self.dirs["input"] = input_dir or (root / "scan")

        # `.scan` directory is used to write intermediate files.
        # Directory tree:
        # .scan/
        # .scan/pic -> pictures extracted from the pdf
        # .scan/cfg/more_infos.csv -> missing students names.
        # .scan/cfg/verified.csv -> pages already verified.
        # .scan/cfg/skipped.csv -> pages to skip.
        # .scan/scores.csv
        # .scan/data -> data stored as .scandata files (used to resume interrupted scan).
        self.dirs["output"] = output_dir or (self.dirs["root"] / ".scan")
        self.dirs["data"] = self.dirs["output"] / "data"
        self.dirs["cfg"] = self.dirs["output"] / "cfg"
        self.dirs["pic"] = self.dirs["output"] / "pic"
        self.dirs["pdf"] = self.dirs["output"] / "pdf"
        self.dirs["log"] = self.dirs["output"] / "log"

        self.files["base"] = search_by_extension(self.dirs["root"], ".ptyx").stem
        self.files["results"] = self.dirs["pdf"] / (self.files["base"] + "-results")

    #    def _extract_pictures_from_pdf(self):
    #        # First, we must test if pdf files have changed.
    #        # - calculate hash for each file.
    #        # - if nothing has changes, pass.
    #        # - if new pdf files are found, extract pictures from them and store their hash.
    #        # - if pdf files where modified or removed, regenerate everything.
    #        # hashlib.sha512(f.read()).hexdigest()

    def _load_all_info(self) -> None:
        """Load all information from files."""
        self._load_data()
        # Read manually entered information (if any).
        self.files["cfg"] = self.dirs["cfg"] / "more_infos.csv"
        if self.files["cfg"].is_file():
            with open(self.files["cfg"], "r", newline="") as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    try:
                        sheet_ID, name, student_ID = row
                    except ValueError:
                        sheet_ID, name = row
                        student_ID = ""
                    self.more_infos[int(sheet_ID)] = (name, student_ID)
                print("Retrieved infos:", self.more_infos)

        # List manually verified pages.
        # They should not be verified anymore.
        self.files["verified"] = self.dirs["cfg"] / "verified.txt"
        if self.files["verified"].is_file():
            with open(self.files["verified"], "r", encoding="utf8", newline="") as file:
                self.verified = set(Path(line.strip()) for line in file.readlines())
                print("Pages manually verified:")
                for path in self.verified:
                    print(f"    • {path}")

        self.skipped = set(pic_names_iterator(self.data))

        # List skipped pictures.
        # Next time, they will be skipped with no warning.
        self.files["skipped"] = self.dirs["cfg"] / "skipped.txt"
        if self.files["skipped"].is_file():
            with open(self.files["skipped"], "r", encoding="utf8", newline="") as file:
                self.skipped |= set(Path(line.strip()) for line in file.readlines())
                print("Pictures skipped:")
                for path in sorted(self.skipped):
                    print(f"    • {path}")

    def scan_picture(self, picture: Union[str, Path], manual_verification: bool = True) -> None:
        """This is used for debuging (it allows to test pages one by one)."""
        # f1-pic-003.jpg (page 25)
        # f12-pic-005.jpg
        # f12-pic-003.jpg
        # f12-pic-004.jpg
        # f12-pic-013.jpg
        # f7-pic-013.jpg
        # f9-pic-004.jpg
        # f9-pic-005.jpg
        # f13-pic-002.jpg
        if not any(str(picture).endswith(ext) for ext in PIC_EXTS):
            raise TypeError("Allowed picture extensions: " + ", ".join(PIC_EXTS))
        pic_path = Path(picture).expanduser().resolve()
        if not pic_path.is_file():
            pic_path = self.dirs["pic"] / picture
        pic_data, _ = scan_picture(pic_path, self.config, manual_verification=manual_verification)
        print(pic_data)

    def _generate_current_pdf_hashes(self) -> dict[str, Path]:
        """Return the hashes of all the pdf files found in `scan/` directory.

        Return: {hash: pdf path}
        """
        hashes = dict()
        for path in self.dirs["input"].glob("**/*.pdf"):
            with open(path, "rb") as pdf_file:
                hashes[blake2b(pdf_file.read(), digest_size=20).hexdigest()] = path
        return hashes

    def _update_input_data(self) -> None:
        """Test if input data has changed, and update it if needed."""
        hash2pdf: dict[str, Path] = self._generate_current_pdf_hashes()

        self._remove_obsolete_files(hash2pdf)

        # For each new pdf files, extract all pictures
        to_extract: list[tuple[Path, Path]] = []
        with Pool() as pool:
            for pdfhash, pdfpath in hash2pdf.items():
                folder = self.dirs["pic"] / pdfhash
                if not folder.is_dir():
                    to_extract.append((pdfpath, folder))
                elif number_of_pages(pdfpath) != len(
                    [f for f in folder.iterdir() if f.suffix.lower() in PIC_EXTS]
                ):
                    # Extraction was probably interrupted
                    rmtree(folder)
                    folder.mkdir()
                    to_extract.append((pdfpath, folder))
            pool.starmap(extract_pdf_pictures, to_extract)

    def _remove_obsolete_files(self, hash2pdf):
        """For each removed pdf files, remove corresponding pictures and data."""
        for path in self.dirs["pic"].iterdir():
            if not path.is_dir():
                raise RuntimeError(
                    f'Folder "{path.parent}" should only contain folders.\n'
                    "You may clean it manually, or remove it with following command:\n"
                    f'rm -r "{path.parent}"'
                )
            if path.name not in hash2pdf:
                rmtree(path)
        for path in self.dirs["data"].glob("*.index"):
            if path.stem not in hash2pdf:
                with open(path) as f:
                    doc_ids = set(f.read().split())
                for doc_id in doc_ids:
                    (self.dirs["data"] / f"{doc_id}.scandata").unlink(missing_ok=True)
                    for webp in self.dirs["data"].glob(f"{doc_id}-*.webp"):
                        webp.unlink()
                path.unlink()

    def _warn(self, *values, sep=" ", end="\n") -> None:
        """Print to stdout and write to log file."""
        msg = sep.join(str(val) for val in values) + end
        print(msg)
        with open(self.dirs["log"] / self.logname, "a", encoding="utf8") as logfile:
            logfile.write(msg)
        self.warnings = True

    def scan_all(
        self,
        start: int = 1,
        end: Union[int, float] = inf,
        manual_verification: Optional[bool] = None,
        ask_for_name: bool = False,
        reset: bool = False,
    ) -> None:
        """Extract information from pdf, calculate scores and annotate documents
        to display correct answers."""
        self._make_dirs(reset)

        # Test if the PDF files of the input directory have changed and
        # extract the images from the PDF files if needed.
        print("Search for previous data...")
        self._update_input_data()
        # Load data from previous run
        self._load_all_info()

        # Dict `data` will collect data from all scanned tests.
        # ...............................................................
        # FORMAT: {ID: {'pages': (dict) the pages seen, and all related information,
        #               'answers': ({int: set}) the answers of the student for each question,
        #               'score': (float) the test score,
        #               'name': (str) the student name,
        #               'last_pic': (str) last image seen full path,
        #               },
        #           ...
        #          }
        # ...............................................................
        #

        # ---------------------------------------
        # Extract informations from the pictures.
        # ---------------------------------------
        data = self.data
        self.name2docID = {d["name"]: ID for ID, d in data.items()}
        self.already_seen = set((ID, p) for ID, d in data.items() for p in d["pages"])
        pic_list = sorted(f for f in self.dirs["pic"].glob("*/*") if f.suffix.lower() in PIC_EXTS)

        assert all(isinstance(path, Path) for path in self.skipped)
        assert all(isinstance(path, Path) for path in self.verified)

        for i, pic_path in enumerate(pic_list, start=1):
            if not (start <= i <= end):
                continue
            # Make pic_path relative, so that folder may be moved if needed.
            pic_path = pic_path.relative_to(self.dirs["pic"])
            if pic_path in self.skipped:
                continue
            print("-------------------------------------------------------")
            print("Page", i)
            print("File:", pic_path)

            # 1) Extract all the data of an image
            #    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾

            try:
                # Warning: manual_verification can be None, so the order is important
                # below : False and None -> False (but None and False -> None).
                manual_verification = (pic_path not in self.verified) and manual_verification
                pic_data, matrix = scan_picture(self.dirs["pic"] / pic_path, self.config, manual_verification)
                # `pic_data` FORMAT is specified in `scan_pic.py`.
                # (Search for `pic_data =` in `scan_pic.py`).
                pic_data["pic_path"] = str(pic_path)
                print()

            except CalibrationError:
                self._warn(f"WARNING: {pic_path} seems invalid ! Skipping...")
                input("-- PAUSE --")
                with open(self.files["skipped"], "a", newline="", encoding="utf8") as file:
                    file.write(f"{pic_path}\n")
                continue

            if pic_data["verified"]:
                # If the page has been manually verified, keep track of it,
                # so it won't be verified next time if a second pass is needed.
                with open(self.files["verified"], "a", newline="", encoding="utf8") as file:
                    file.write(f"{pic_path}\n")

            doc_id = pic_data["ID"]
            page = pic_data["page"]

            if self._keep_previous_version(pic_data):
                continue

            # 2) Gather data
            #    ‾‾‾‾‾‾‾‾‾‾‾
            name, student_ID = self.more_infos.get(doc_id, ("", ""))
            doc_data: DocumentData = data.setdefault(
                doc_id,
                DocumentData(
                    pages={}, name=name, student_ID=student_ID, answered={}, score=0, score_per_question={}
                ),
            )
            doc_data["pages"][page] = pic_data

            for q in pic_data["answered"]:
                ans = doc_data["answered"].setdefault(q, set())
                ans |= pic_data["answered"][q]
                # Simplify: doc_data["answered"][q] = set(pic_data["answered"][q])

            # 3) 1st page of the test => retrieve the student name
            #    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
            if page == 1:
                self._extract_name(doc_id, doc_data, matrix)

            # Store work in progress, so we can resume process if something fails...
            self._store_data(pic_path.parent.name, doc_id, page, matrix)

        # ---------------------------
        # Test integrity
        # ---------------------------
        # For every test:
        # - all pages must have been scanned,
        # - all questions must have been seen.
        print("Test for data integrity...")
        self._test_integrity()
        print("Everything seems OK.")

        # ---------------------------
        # Calculate scores
        # ---------------------------
        # Calculate the score, taking care of the chosen mode.
        self._calculate_scores()
        print()

        # ---------------------------------------------------
        # Time to synthetize & store all those informations !
        # ---------------------------------------------------
        self.generate_output()

    def resolve_conflicts(self):
        """Resolve conflicts manually: unknown student ID, ambiguous answer..."""
        pass


def scan(
    path: Path,
    reset: bool = False,
    ask_for_name: bool = False,
    verify: Literal["auto", "always", "never"] = "auto",
) -> None:
    """Implement `mcq scan` command."""
    if verify == "always":
        manual_verification: Optional[bool] = True
    elif verify == "never":
        manual_verification = False
    else:
        manual_verification = None
    MCQPictureParser(path).scan_all(
        reset=reset,
        ask_for_name=ask_for_name,
        manual_verification=manual_verification,
    )



from json import loads, dumps as _dumps, JSONEncoder
from pathlib import Path
from typing import Dict, Set, Any, Tuple, Optional, Union, TypeVar, TypedDict, List, Literal

T = TypeVar("T")


class OrderingConfiguration(TypedDict):
    questions: List[int]
    answers: Dict[int, List[Tuple[int, bool | None]]]


QuestionNumberOrDefault = Union[Literal["default"], int]

StudentIdFormat = Tuple[int, int, List[Tuple[str, ...]]]


# TODO: improve typing precision
class Configuration(TypedDict, total=False):
    ordering: Dict[int, OrderingConfiguration]
    mode: Dict[QuestionNumberOrDefault, str]
    correct: Dict[QuestionNumberOrDefault, float]
    incorrect: Dict[QuestionNumberOrDefault, float]
    skipped: Dict[QuestionNumberOrDefault, float]
    # -inf and inf would be sensible defaults for floor and ceil, but unfortunately they aren't supported
    # by ast.literal_eval().
    floor: Dict[QuestionNumberOrDefault, float | None]
    ceil: Dict[QuestionNumberOrDefault, float | None]
    students: List[str]
    id_table_pos: Tuple[float, float]
    id_format: StudentIdFormat
    students_ids: Dict[str, str]
    students_list: List[str]
    boxes: Dict[int, Dict[int, Dict[str, Tuple[float, float]]]]
    max_score: float


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        # Save sets as tuples.
        if isinstance(obj, set):
            return tuple(obj)
        # Let the base class default method raise the TypeError
        return JSONEncoder.default(self, obj)


def dumps(o: Any) -> str:
    return _dumps(o, ensure_ascii=False, cls=CustomJSONEncoder)


def fmt(s: str) -> str:
    return s.upper().center(40, "-")


def encode2js(o, formatter=(lambda s: s), _level=0) -> str:
    """Encode dict `o` to JSON.

    If `formatter` is set, it must be a function used to format first-level
    keys of `o`."""
    if _level == 0 and not isinstance(o, dict):
        raise NotImplementedError
    if isinstance(o, dict):
        if _level == 0:
            blocks = []
            for k, v in o.items():
                blocks.append(f'"{formatter(k)}":\n' + encode2js(v, _level=_level + 1))
            return "{\n\n%s\n\n}" % ",\n\n".join(blocks)
        else:
            blocks = []
            indent = (_level - 1) * "\t"
            for k, v in o.items():
                # Warning: all JSON keys must be strings !
                blocks.append(f'{indent}"{k}": ' + encode2js(v, _level=_level + 1))
            return "{\n%s\n%s}" % (",\n".join(blocks), indent)
    elif isinstance(o, (tuple, set, list)):
        assert _level != 0
        if _level == 1:
            return "[\n%s\n]" % ",\n".join(dumps(v) for v in o)
        else:
            return dumps(o)
    else:
        return dumps(o)


def keys2int(d: Dict[str, T]) -> Dict[Union[int, str], T]:
    return {(int(k) if k.isdecimal() else k): v for k, v in d.items()}


def decodejs(js: str) -> Configuration:
    d = loads(js, object_hook=keys2int)
    new_d: Configuration = {}
    # Strip '-' from keys and convert them to lower case.
    for key in d:
        new_key = key.strip("-").lower()
        assert new_key in Configuration.__annotations__
        new_d[new_key] = d[key]  # type: ignore
    # Students ID must be kept as strings.
    new_d["students_ids"] = {str(key): val for key, val in new_d["students_ids"].items()}
    return new_d


def dump(path: Union[Path, str], cfg: Configuration) -> None:
    """Dump `cfg` dict to `path` as json."""
    with open(path, "w") as f:
        f.write(encode2js(cfg, formatter=fmt))


def load(path: Union[Path, str]) -> Configuration:
    """Load `path` configuration file (json) and return a dict."""
    with open(path) as f:
        js = f.read()
    return decodejs(js)


def real2apparent(
    original_q_num: int, original_a_num: Optional[int], config: Configuration, doc_id: int
) -> Tuple[int, Optional[int]]:
    """Return apparent question number and answer number.

    If `a` is None, return only question number.

    By "apparent", it means question and answer numbers as they
    will appear in the PDF file, after shuffling questions and answers.

    Arguments `q` and `a` are real question and answer numbers, that is
    the ones before questions and answers were shuffled."""
    questions = config["ordering"][doc_id]["questions"]
    answers = config["ordering"][doc_id]["answers"]
    # Apparent question number (ie. after shuffling).
    # Attention, list index 0 corresponds to question number 1.
    pdf_q_num = questions.index(original_q_num) + 1
    if original_a_num is None:
        return pdf_q_num, None
    for pdf_a_num, (ans_num, is_correct) in enumerate(answers[original_q_num], start=1):
        if ans_num == original_a_num:
            return pdf_q_num, pdf_a_num
    raise IndexError(f"Answer {original_a_num} not found for question {original_q_num}.")


def apparent2real(
    pdf_q_num: int, pdf_a_num: Optional[int], config: dict, doc_id: int
) -> Tuple[int, Optional[int]]:
    """Return real question number and answer number.

    If `a` is None, return only question number.
    """
    questions = config["ordering"][doc_id]["questions"]
    answers = config["ordering"][doc_id]["answers"]
    # Real question number (ie. before shuffling).
    # Attention, first question is numbered 1, corresponding to list index 0.
    original_q_num = questions[pdf_q_num - 1]
    if pdf_a_num is None:
        return original_q_num, None
    # Real answer number (ie. before shuffling).
    original_a_num = answers[original_q_num][pdf_a_num - 1][0]
    return original_q_num, original_a_num


def is_answer_correct(
    q_num: int, a_num: int, config: Configuration, doc_id: int, use_original_num: bool = True
) -> bool | None:
    if use_original_num:
        # q_num and a_num are question and answer number *before* shuffling questions.
        for original_a_num, is_correct in config["ordering"][doc_id]["answers"][q_num]:
            if original_a_num == a_num:
                return is_correct
        raise IndexError(f"Answer {a_num} not found.")
    else:
        # q_num and a_num are question and answer number *after* shuffling questions.
        original_q_num = config["ordering"][doc_id]["questions"][q_num - 1]
        return config["ordering"][doc_id]["answers"][original_q_num][a_num - 1][1]


def get_answers_with_status(
    config: Configuration, *, correct: bool | None, use_original_num: bool = True
) -> Dict[int, Dict[int, Set[int]]]:
    """Return a dict containing the set of the correct answers for each question for each document ID.

    By default, questions and answers are numbered following their original order of definition
    in the ptyx file.

    If `use_original_num` is set to `False`, questions numbers and correct answers numbers returned
    will be apparent ones (i.e. the number that appear in the document, after shuffling).
    """
    correct_answers_by_id = {}
    for doc_id in config["ordering"]:
        correct_answers = {}
        questions = config["ordering"][doc_id]["questions"]
        answers = config["ordering"][doc_id]["answers"]
        for i, q in enumerate(questions, start=1):
            # `i` is the 'apparent' question number.
            # `q` is the 'real' question number, i.e. the question number before shuffling.
            # `j` is the 'apparent' answer number.
            # `a` is the 'real' answer number, i.e. the answer number before shuffling.
            if use_original_num:
                correct_answers[q] = {a for (a, _is_correct) in answers[q] if _is_correct == correct}
            else:
                correct_answers[i] = {
                    j for (j, (a, _is_correct)) in enumerate(answers[q], start=1) if _is_correct == correct
                }
        correct_answers_by_id[doc_id] = correct_answers
    return correct_answers_by_id



import csv
from pathlib import Path
from string import ascii_letters
from typing import Sequence, List, Optional, Dict, Union, Tuple

from ..parameters import (
    CELL_SIZE_IN_CM,
    MARGIN_LEFT_IN_CM,
    MARGIN_RIGHT_IN_CM,
    PAPER_FORMAT,
    MARGIN_BOTTOM_IN_CM,
    MARGIN_TOP_IN_CM,
    CALIBRATION_SQUARE_POSITION,
    CALIBRATION_SQUARE_SIZE,
)
from ..tools.config_parser import get_answers_with_status, Configuration


class IdentifiantError(RuntimeError):
    pass


def _byte_as_codebar(byte: Union[int, str], n=0) -> str:
    """Generate LaTeX code (TikZ) for byte number `n` in ID band.

    - `byte` is a number between 0 and 255, or some LaTeX code resulting
      in such a number.
      If the number is above 255, encoding will be wrong.
    - `n` should be incremented at each call, so as not to overwrite
      a previous codebar.
    """
    if isinstance(byte, int) and byte > 255:
        raise NotImplementedError("Value can't exceed 255 for one byte.")
    return rf"""
    \n={byte};
    \j={n};
    for \i in {{1,...,8}}{{%
            \r = int(Mod(\n,2));
            \n = int(\n/2);
            {{\draw[fill=color\r] ({{\i*0.25+2*\j}},0) rectangle ({{\i*0.25+0.25+2*\j}},0.25);
             }};
            }};"""


def ID_band(doc_id: int, calibration=True) -> str:
    """Generate top banner of the page to be scanned later.

    `doc_id` is the integer which identifies each document.

    This top banner is made of:
    - four squares for calibration (top left, top right,
      bottom left, bottom right, at 1 cm of the border).
      If `calibration` is False, those squares are not displayed.

    - an identification band (kind of codebar) midway between
      the top squares.
      This ID band is used to encode both the MCQ ID number and the page
      number. They will be read automatically later.

    NB:
    - There is no need to provide the page number, since it will be
      handled directly by LaTeX.
    - Page number can't exceed 255.
    - Document number can't exceed 65535.
    """
    if doc_id >= 2**16:
        raise ValueError(f"Document number can't exceed 65535 (current id: {doc_id}).")
    latex = [
        r"\newcommand\CustomHeader{%"
        "\n    "
        r"\begin{tikzpicture}[remember picture,overlay,black,"  # Set default color to black !
        r"every node/.style={inner sep=0,outer sep=-0.2}]"
    ]
    if calibration:
        pos = CALIBRATION_SQUARE_POSITION
        pos2 = CALIBRATION_SQUARE_POSITION + CALIBRATION_SQUARE_SIZE
        latex.append(
            rf"""
        \draw[fill=black] ([xshift={pos}cm,yshift=-{pos}cm]current page.north west)
            rectangle ([xshift={pos2}cm,yshift=-{pos2}cm]current page.north west);
        \draw[fill=black] ([xshift=-{pos}cm,yshift=-{pos}cm]current page.north east)
            rectangle ([xshift=-{pos2}cm,yshift=-{pos2}cm]current page.north east);
        \draw[fill=black] ([xshift={pos}cm,yshift={pos}cm]current page.south west)
            rectangle ([xshift={pos2}cm,yshift={pos2}cm]current page.south west);
        \draw[fill=black] ([xshift=-{pos}cm,yshift={pos}cm]current page.south east)
            rectangle ([xshift=-{pos2}cm,yshift={pos2}cm]current page.south east);"""
        )
    latex.append(
        r"""\node at ([yshift=-1cm]current page.north) [anchor=north] {
            \begin{tikzpicture}
            \definecolor{color0}{rgb}{1,1,1}
            \definecolor{color1}{rgb}{0,0,0}
            \draw[fill=black] (0,0) rectangle (0.25,0.25);
            \tikzmath {"""
    )
    latex.append(_byte_as_codebar(r"\thepage"))
    latex.append(_byte_as_codebar(doc_id % 256, n=1))
    latex.append(_byte_as_codebar(doc_id // 256, n=2))
    latex.append(
        rf"""}}
        \node[anchor=west] at  ({{2.5+2*\j}},0.1)
            {{\scriptsize\textbf{{\#{doc_id}}}~:~{{\thepage}}/\zpageref{{LastPage}}}};
        \end{{tikzpicture}}}};

        \draw[dotted]  ([xshift=-1cm,yshift=-2cm]current page.north east)
            -- ([xshift=1cm,yshift=-2cm]current page.north west)
            node [pos=0.25,fill=white]
            {{\,\,\scriptsize\textuparrow\,\,\textsc{{N'écrivez rien au
            dessus de cette ligne}}\,\,\textuparrow\,\,}}
            node [pos=0.75,fill=white]
            {{\,\,\scriptsize\textuparrow\,\,\textsc{{N'écrivez rien au
            dessus de cette ligne}}\,\,\textuparrow\,\,}};
    \end{{tikzpicture}}}}"""
    )
    return "".join(latex)


def extract_students_id_and_name_from_csv(csv_path: Path, script_path: Path) -> Dict[str, str]:
    """`csv_path` is the path of the CSV file who contains students names and ids.
    The first column of the CSV file must contain the ids.

    Return a dictionnary containing the students ID and corresponding names.
    """
    csv_path = csv_path.expanduser()
    if not csv_path.is_absolute():
        csv_path = (script_path.parent / csv_path).absolute()
    # XXX: support ODS and XLS files ?
    # soffice --convert-to cvs filename.ods
    # https://ask.libreoffice.org/en/question/2641/convert-to-command-line-parameter/
    ids: Dict[str, str] = {}
    # Read CSV file and generate the dictionary {id: "student name"}.
    with open(csv_path) as f:
        dialect = csv.Sniffer().sniff(f.read(1024))
        f.seek(0)
        for row in csv.reader(f, dialect):
            id_, *row = row
            id_ = id_.strip()
            name = " ".join(item.strip() for item in row)
            if id_ in ids and ids[id_] != name:
                raise RuntimeError(f"Error: same ID {id_!r} for different students in {csv_path!r} !")
            ids[id_] = name
    return ids


def extract_students_name_from_csv(csv_path: Path, script_path: Path) -> List[str]:
    """`csv_path` is the path of the CSV file who contains students names.

    Return a list of students names.
    """
    csv_path = csv_path.expanduser()
    if not csv_path.is_absolute():
        csv_path = (script_path.parent / csv_path).absolute()

    names: List[str] = []
    # Read CSV file and generate the dictionary {id: "student name"}.
    with open(csv_path) as f:
        for row in csv.reader(f):
            names.append(" ".join(item.strip() for item in row))
    return names


def students_checkboxes(names: Sequence[str], _n_student=None) -> str:
    """Generate a list of all students, where student can check his name.

    `names` is a list of students names.
    `_n_student` is used to prefilled the table (for debuging).
    """
    content = [
        r"""
        \vspace{-1em}
        \begin{center}
        \begin{tikzpicture}[scale=.25]
        \draw [fill=black] (-2,0) rectangle (-1,1) (-1.5,0) node[below]
        {\tiny\rotatebox{-90}{\texttt{\textbf{Noircir la case}}}};"""
    ]

    # Generate the corresponding names' table in LaTeX.
    b = None
    for i, name in enumerate(reversed(names)):
        # Truncate long names.
        if len(name) >= 15:
            _name = name[:13].strip()
            if " " not in name[12:13]:
                _name += "."
            name = _name
        a = 2 * i
        b = a + 1
        c = a + 0.5
        color = "black" if _n_student == len(names) - i else "white"
        content.append(
            rf"""\draw[fill={color}] ({a},0) rectangle ({b},1) ({c},0)
            node[below] {{\tiny \rotatebox{{-90}}{{\texttt{{{name}}}}}}};"""
        )
    if b is None:
        # No names.
        return ""
    b += 1
    content.append(
        rf"""\draw[rounded corners] (-3,2) rectangle ({b}, -6.5);
        \draw[] (-0.5,2) -- (-0.5,-6.5);
        \end{{tikzpicture}}
        \end{{center}}
        \vspace{{-1em}}
        """
    )
    return "\n".join(content)


def student_id_table(ID_length: int, max_ndigits: int, digits: List[Tuple[str, ...]]) -> str:
    """Generate a table where the student will write its identification number.

    The table have a row for each digit, where the student check corresponding
    digit to indicate its number.

    Parameters:
    - the length of an ID,
    - the maximal number of different digits in an ID caracter,
    - a list of sets corresponding to the different digits used for each ID caracter.

    Return: LaTeX code.
    """
    content: List[str] = []
    write = content.append
    write("\n\n")
    write(r"\begin{tikzpicture}[baseline=-10pt,scale=.5]")
    write(r"\node[anchor=south west] at (-1, 0) {Numéro étudiant (INE)~:};")
    write(r"\draw[] (-1, 0) node {\zsavepos{ID-table}} rectangle (0,%s);" % (-ID_length))
    for j in range(ID_length):
        # One row for each digit of the student id number.
        for i, d in enumerate(sorted(digits[j])):
            write(
                rf"""\draw ({i},{-j}) rectangle ({i+1},{-j-1})
                    ({i+0.25},{-j-0.25}) node  {{\footnotesize\color{{black}}\textsf{{{d}}}}};"""
            )
        for i in range(i, max_ndigits):
            write(rf"""\draw ({i},{-j}) rectangle ({i+1},{-j-1});""")
    write(r"\draw[black,->,thick] (-0.5, -0.5) -- (-0.5,%s);" % (0.5 - ID_length))
    write(r"\end{tikzpicture}")
    write(
        r"\hfill\begin{tikzpicture}[baseline=10pt]"
        r"\node[draw,rounded corners] {\begin{tabular}{p{8cm}}"
        r"\textsc{Nom~:}~\dotfill\\"
        r"Prénom~:~\dotfill\\"
        r"Groupe~:~\dotfill\\"
        r"Numéro d'étudiant:~\dotfill\\"
        r"\end{tabular}};\end{tikzpicture}"
        r"\write\mywrite{ID-table: "
        r"(\dimtomm{\zposx{ID-table}sp}, "
        r"\dimtomm{\zposy{ID-table}sp})}"
    )
    write("\n\n")
    return "\n".join(content)


def table_for_answers(config: Configuration, doc_id: Optional[int] = None) -> str:
    """Generate the table where students select correct answers.

    - `config` is a dict generated when compiling test.
    - `ID` is the test ID if correct answers should be shown.
      If `ID` is `None` (default), the table will be blank.
    """
    content: List[str] = []
    write = content.append

    # Generate the table where students will answer.
    tkzoptions = ["scale=%s" % CELL_SIZE_IN_CM]

    d = config["ordering"][1 if doc_id is None else doc_id]
    questions = d["questions"]
    answers = d["answers"]
    n_questions = len(questions)
    n_max_answers = max(len(nums) for nums in answers.values())
    flip = n_max_answers > n_questions
    if flip:
        tkzoptions.extend(["x={(0cm,-1cm)}", "y={(-1cm,0cm)}"])

    write(
        r"""
        \begin{tikzpicture}[%s]
        \draw[thin,fill=black] (-1,0) rectangle (0,1);"""
        % (",".join(tkzoptions))
    )

    for x1 in range(n_questions):
        x2 = x1 + 1
        x3 = 0.5 * (x1 + x2)
        write(rf"\draw[ultra thin] ({x1},0) rectangle ({x2},1) ({x3},0.5) " rf"node {{{x1 + 1}}};")

    # Find correct answers numbers for each question.
    if doc_id is not None:
        correct_ans = get_answers_with_status(config, correct=True, use_original_num=False)[doc_id]

    # i = -1
    for i in range(n_max_answers):
        name = ascii_letters[i]
        y1 = -i
        y2 = y1 - 1
        y3 = 0.5 * (y1 + y2)
        write("\n" rf"\draw[ultra thin] (-1,{y1}) rectangle (0,{y2}) (-0.5,{y3}) " rf"node {{{name}}};")
        for j in range(n_questions):
            x1 = j
            x2 = x1 + 1
            opt = ""
            if doc_id is not None and i + 1 in correct_ans[j + 1]:
                opt = "fill=gray"
            write(rf"\draw [ultra thin,{opt}] ({x1},{y1}) rectangle ({x2},{y2});")

    write(rf"\draw [thick] (-1,1) rectangle ({x2},{y2});" "\n")
    #              %\draw [thick] (-1,0) -- ({x2},0);

    for i in range(0, x2):
        write(rf"\draw [thick] ({i},1) -- ({i},{y2});" "\n")

    write(r"\end{tikzpicture}\hfill\hfill\hfil" "\n")

    return "\n".join(content)


def packages_and_macros() -> tuple[str, str]:
    """Generate LaTeX default header (loading LaTeX packages and defining some custom macros)."""
    # https://tex.stackexchange.com/questions/37297/how-to-get-element-position-in-latex
    paper_format = f"{PAPER_FORMAT.lower()}paper"
    # LaTeX header is in two part, so as user may insert some customization here.
    return (
        rf"""\documentclass[{paper_format},twoside,10pt]{{article}}
    \PassOptionsToPackage{{utf8}}{{inputenc}}
    \PassOptionsToPackage{{document}}{{ragged2e}}
    \PassOptionsToPackage{{left={MARGIN_LEFT_IN_CM}cm,
        right={MARGIN_RIGHT_IN_CM}cm,
        top={MARGIN_TOP_IN_CM}cm,bottom={MARGIN_BOTTOM_IN_CM}cm}}{{geometry}}
    \parindent=0cm
    \newcommand*\graysquared[1]{{\tikz[baseline=(char.base)]{{
        \node[fill=gray,shape=rectangle,draw,inner sep=2pt] (char) {{\color{{white}}\textbf{{#1}}}};}}}}
    \newcommand*\whitesquared[1]{{\tikz[baseline=(char.base)]{{
        \node[fill=white,shape=rectangle,draw,inner sep=2pt] (char) {{\color{{black}}\textbf{{#1}}}};}}}}
    \newcommand*\ptyxMCQcircled[1]{{\tikz[baseline=(char.base)]{{
        \node[shape=circle,fill=blue!20!white,draw,inner sep=2pt] (char) {{\textbf{{#1}}}};}}}}
    \makeatletter
    \newcommand{{\ptyxMCQsimfill}}{{%
    \leavevmode \cleaders \hb@xt@ .50em{{\hss $\sim$\hss }}\hfill \kern \z@
    }}
    \makeatother
    \newcounter{{answerNumber}}
    \renewcommand{{\thesubsection}}{{\Alph{{subsection}}}}
    """,
        # <Custom packages will be loaded just here.>
        r"""\usepackage{inputenc}
    \usepackage{ragged2e}
    \usepackage{geometry}
    \usepackage{pifont}
    \usepackage{textcomp}
    \usepackage{nopageno}
    \usepackage{tikz}
    \usepackage{zref-user}
    \usepackage{zref-abspos}
    \usepackage{zref-abspage}
    \usepackage{zref-lastpage}
    \usepackage{everypage}
    \usepackage{tabularx}
    \usetikzlibrary{calc}
    \usetikzlibrary{math}
    \makeatletter
    \newcommand\dimtomm[1]{%
        \strip@pt\dimexpr 0.351459804\dimexpr#1\relax\relax%
    }
    \makeatother
    \newcommand{\checkBox}[2]{%
        \begin{tikzpicture}[baseline=-12pt,color=black, thick]
            \draw[fill=#1] (0,0)
                node {\zsavepos{#2-ll}}
                rectangle (.5,-.5);
        \end{tikzpicture}%
        \write\mywrite{#2: p\thepage, (%
            \dimtomm{\zposx{#2-ll}sp},
            \dimtomm{\zposy{#2-ll}sp})%
        }%
    }
    \newwrite\mywrite
    \openout\mywrite=\jobname.pos\relax
    \usepackage{enumitem} % To resume an enumeration.
    \setenumerate[0]{label=\protect\ptyxMCQcircled{\arabic*}}
    \AddEverypageHook{\CustomHeader}

    \newlength{\ptyxMCQTabLength}
    \newcommand{\ptyxMCQTab}[2]{%
      \settowidth{\ptyxMCQTabLength}{#1{}#2}
      \ifdim \ptyxMCQTabLength<\textwidth%
      \begin{tabular}{l@{\,\,}l}#1&#2\end{tabular}%
      \else%
      \begin{tabularx}{\linewidth}{l@{\,\,}X}#1&#2\end{tabularx}%
      \fi%
    }
    """,
    )


def answers_and_score(config: Configuration, name: str, identifier: int, score: Optional[float]) -> str:
    """Generate plain LaTeX code corresponding to score and correct answers."""
    table = table_for_answers(config, identifier)
    if score is not None:
        max_score = config["max_score"]
        score_latex = (
            r"""\begin{tikzpicture}
            \node[draw,very thick,rectangle, rounded corners,red!70!black] (0,0) {
            \begin{Large}
            Score~: %(score)s/%(max_score)s
            \end{Large}};
            \end{tikzpicture}"""
            % locals()
        )
    else:
        score_latex = ""
    left = MARGIN_LEFT_IN_CM
    right = MARGIN_RIGHT_IN_CM
    top = MARGIN_TOP_IN_CM
    bottom = MARGIN_BOTTOM_IN_CM
    paper = PAPER_FORMAT
    return (
        r"""
    \documentclass[%(paper)s,10pt]{article}
    \usepackage[utf8]{inputenc}
    \usepackage[document]{ragged2e}
    \usepackage{nopageno}
    \usepackage{tikz}
    \usepackage[left=%(left)scm,right=%(right)scm,top=%(top)scm,bottom=%(bottom)scm]{geometry}
    \parindent=0cm
    \usepackage{textcomp}

    \begin{document}
    \begin{Large}\textsc{%(name)s}\end{Large}
    \hfill%(score_latex)s

    \bigskip

    Solution~:
    \medskip

    %(table)s

    \end{document}
    """
        % locals()
    )



"""
Generate pdf file from raw mcq file.
"""
import sys
import traceback
from pathlib import Path

from ptyx.compilation import make_files, make_file
from ptyx.latex_generator import compiler, Compiler

from ..tools.config_parser import dump, Configuration


def generate_config_file(_compiler: Compiler) -> None:
    mcq_data: Configuration = _compiler.latex_generator.mcq_data
    file_path = _compiler.file_path
    assert file_path is not None
    folder = file_path.parent
    name = file_path.stem
    id_table_pos = None
    for n in mcq_data["ordering"]:
        # XXX: what if files are not auto-numbered, but a list
        # of names is provided to Ptyx instead ?
        # (cf. command line options).
        if len(mcq_data["ordering"]) == 1:
            filename = f"{name}.pos"
        else:
            filename = f"{name}-{n}.pos"
        full_path = folder / ".compile" / name / filename
        # For each page of the document, give the position of every answer's checkbox.
        checkboxes_positions: dict[int, dict[str, tuple[float, float]]] = {}
        mcq_data["boxes"][n] = checkboxes_positions
        with open(full_path) as f:
            for line in f:
                k, v = line.split(": ", 1)
                k = k.strip()
                if k == "ID-table":
                    if id_table_pos is None:
                        id_table_pos = tuple(float(s.strip("() \n")) for s in v.split(","))
                        assert len(id_table_pos) == 2
                        mcq_data["id_table_pos"] = id_table_pos  # type: ignore
                    continue
                page, x, y = [s.strip("p() \n") for s in v.split(",")]
                checkboxes_positions.setdefault(int(page), {})[k] = (float(x), float(y))

    config_file = file_path.with_suffix(".ptyx.mcq.config.json")
    dump(config_file, mcq_data)


def get_ptyxfile_path(path: Path) -> Path:
    """Get the path of the ptyx file corresponding to the given `path`.

    If `path` is already the path of a ptyx file, just return `path` unchanged.
    Else, `path` must be a directory which contains a single ptyx file, and the path
    of this ptyx file is returned.

    Raise `FileNotFoundError` if `path` is neither a directory nor a ptyx file, or if `path`
    is a directory but contains no ptyx file or several ptyx files.
    """
    path = path.resolve()
    if path.suffix == ".ptyx":
        if not path.is_file():
            raise FileNotFoundError(f"File '{path}' not found.")
    else:
        all_ptyx_files: list[Path] = list(path.glob("*.ptyx"))
        if len(all_ptyx_files) == 0:
            raise FileNotFoundError(f"No .ptyx file found in '{path}'.")
        elif len(all_ptyx_files) > 1:
            raise FileNotFoundError(f"Several .ptyx file found in '{path}', I don't know which one to chose.")
        path = all_ptyx_files[0]
    return path


def make(
    path: Path, num: int = 1, start: int = 1, quiet: bool = False, correction_only: bool = False
) -> None:
    """Wrapper for _make(), so that `argparse` module don't intercept exceptions."""
    try:
        _make(path, num, start, quiet, correction_only)
        print(
            "\n\u001b[32;1mCongratulations ! Document was successfully generated "
            f"in {num} version(s).\u001b[0m"
        )
    except Exception as e:  # noqa
        if hasattr(e, "msg"):
            traceback.print_tb(e.__traceback__)
            print(e.msg)  # type: ignore
            print(f"\u001b[31m{e.__class__.__name__}:\u001b[0m {e}")
        else:
            traceback.print_exc()
        print("\n\u001b[31;1mERROR: `mcq make` failed to compile document (see above for details).\u001b[0m")
        sys.exit(1)


def _make(
    path: Path, num: int = 1, start: int = 1, quiet: bool = False, correction_only: bool = False
) -> None:
    """Implement `mcq make` command.

    If `only_correction` is `True`, only generate correction (useful for fast testing).
    """
    assert isinstance(num, int)
    ptyx_filename = get_ptyxfile_path(path)
    # Read pTyX file.
    print(f"Reading {ptyx_filename}...")
    compiler.read_file(ptyx_filename)
    # Parse #INCLUDE tags, load extensions if needed, read seed.
    compiler.preparse()

    # Generate syntax tree.
    # The syntax tree is generated only once, and will then be used
    # for all the following compilations.
    compiler.generate_syntax_tree()

    # Compile and generate output files (tex or pdf)
    output_name, nums = make_files(
        ptyx_filename,
        correction=correction_only,
        compress=True,
        number_of_documents=num,
        fixed_number_of_pages=True,
        quiet=quiet,
        start=start,
        # cpu_cores=1,
    )
    generate_config_file(compiler)

    # Keep track of the seed used.
    seed_value = compiler.seed
    seed_file_name = output_name.parent / ".seed"
    with open(seed_file_name, "w") as seed_file:
        seed_file.write(str(seed_value))

    if not correction_only:
        _, nums2 = make_files(ptyx_filename, correction=True, _nums=nums, compress=True, quiet=quiet)
        assert nums2 == nums, repr((nums, nums2))

        # Generate a document including the different versions of all the questions.
        make_file(
            (output_name.parent / output_name.stem).with_suffix(".all.pdf"),
            context={"MCQ_KEEP_ALL_VERSIONS": True},
            quiet=quiet,
        )
        # Generate a document including the different versions of all the questions
        # with the correct answers checked.
        make_file(
            (output_name.parent / output_name.stem).with_suffix(".all-corr.pdf"),
            context={"MCQ_KEEP_ALL_VERSIONS": True, "PTYX_WITH_ANSWERS": True},
            quiet=quiet,
        )



# -*- coding: utf-8 -*-

######################################
#
#    Scriptlib (python 3+)
#
######################################
#
#    This library provides some shell functions to write easily
#    shell scripts in Python.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
######################################

import glob
import os
import shutil
import subprocess
import sys

_path_7zip = r"C:\Program Files\7-Zip\7z.exe"


def norm(path):
    return os.path.realpath(os.path.normpath(os.path.expanduser(path)))


def cp(src, dst, **kw):
    src_found = False
    for pth in glob.glob(norm(src)):
        if os.path.isfile(pth):
            shutil.copy(pth, norm(dst))
            src_found = True
        elif os.path.isdir(pth):
            shutil.copytree(pth, norm(dst))
            src_found = True
    if not src_found and not kw.get("quiet", False):
        print("Warning: %s not found." % repr(src))


def mv(src, dst):
    for pth in glob.glob(norm(src)):
        shutil.move(pth, norm(dst))


def rm(*paths, **kw):
    quiet = kw.get("quiet", False)
    recursive = kw.get("recursive", False)
    if recursive:
        cwd = os.getcwd()
        listcwd = os.listdir(".")
    for path in paths:
        pths = glob.glob(norm(path))
        if not (pths or quiet or recursive):
            print("Warning: %s not found, couldn't be removed." % path)
        for pth in pths:
            if os.path.isfile(pth):
                os.remove(pth)
        if recursive:
            # noinspection PyUnboundLocalVariable
            for pth in listcwd:
                if os.path.isdir(pth):
                    cd(pth)
                    rm(*paths, **kw)
                    # noinspection PyUnboundLocalVariable
                    cd(cwd)


def rename(src, dst):
    os.rename(norm(src), norm(dst))


def rmdir(*paths, **kw):
    quiet = kw.get("quiet", False)
    for path in paths:
        pths = glob.glob(norm(path))
        if not (pths or quiet):
            print("Warning: %s not found, couldn't be removed." % path)
        for pth in pths:
            if os.path.isdir(pth):
                shutil.rmtree(norm(pth))


def mkdir(path):
    os.mkdir(norm(path))


def cd(path):
    return os.chdir(norm(path))


def command(string, quiet=False):
    out = subprocess.Popen(string, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout
    output = out.read()
    sys.stdout.write(str(output))
    out.close()
    if not quiet:
        print("Commande '%s' executee." % string)
    return output


def zip7(string):
    command('"%s" %s' % (_path_7zip, string))


# ~ def compil(script):
# ~ command(sys.executable + ' -O ' + script + ' py2exe')


def append(srcs, dst):
    with open(dst, "wb") as dest:
        for src in srcs:
            with open(src, "rb") as source:
                dest.write(source.read())


def pause(string="\n-- pause --\n"):
    input(string)


def ls(path="."):
    if "*" in path:
        return glob.glob(path)
    return os.listdir(norm(path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ptyx MCQ Command Line Interface

@author: Nicolas Pourcelot
"""
import shutil
import sys
from argparse import ArgumentParser
from os import unlink

from pathlib import Path
from typing import Optional

from .compile.make import make, get_ptyxfile_path
from .scan.scanner import scan


def main(args: Optional[list] = None) -> None:
    """Main entry point, called whenever `mcq` command is executed."""
    parser = ArgumentParser(description="Generate and manage pdf MCQs.")
    subparsers = parser.add_subparsers()
    add_parser = subparsers.add_parser

    # create the parser for the "new" command
    new_parser = add_parser("new", help="Create an empty ptyx file.")
    new_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default="new-mcq")
    new_parser.set_defaults(func=new)

    # create the parser for the "make" command
    make_parser = add_parser("make", help="Generate pdf file.")
    make_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    make_parser.add_argument(
        "--num",
        "-n",
        metavar="N",
        type=int,
        default=1,
        help="Specify how many versions of the document must be generated.",
    )
    make_parser.add_argument(
        "--start",
        "-s",
        metavar="START",
        type=int,
        default=1,
        help="First document number (default=1).",
    )
    make_parser.add_argument("--quiet", "-q", action="store_true", help="Hide pdflatex output.")
    make_parser.add_argument(
        "--correction-only", "-co", action="store_true", help="Generate only correction for fast testing."
    )
    make_parser.set_defaults(func=make)

    # create the parser for the "scan" command
    scan_parser = add_parser("scan", help="Generate scores from scanned documents.")
    scan_parser.add_argument(
        "path",
        nargs="?",
        default=".",
        help=(
            "Path to a directory which must contain "
            "a .mcq.config file and a .scan.pdf file "
            "(alternatively, this path may point to any file in this folder)."
        ),
    )

    scan_parser.add_argument(
        "--reset",
        action="store_true",
        help="Delete all cached data." "The scanning process will restart from the beginning.",
    )

    scan_parser.add_argument(
        "--verify",
        "--manual-verification",
        choices=("always", "never", "auto"),
        default="auto",
        help="If set to `always`, then for each page scanned, display a picture of "
        "the interpretation by the detection algorithm, "
        "for manual verification.\n"
        "If set to `never`, always assume algorithm is right.\n"
        "Default is `auto`, i.e. only ask for manual verification "
        "in case of ambiguity.",
    )

    scan_parser.add_argument(
        "--ask-for-name",
        action="store_true",
        default=False,
        help="For each first page, display a picture of " "the top of the page and ask for the student name.",
    )

    scan_parser.set_defaults(func=scan)

    # create the parser for the "clear" command
    new_parser = add_parser("clear", help="Remove every MCQ data but the ptyx file.")
    new_parser.add_argument("path", nargs="?", metavar="PATH", type=Path, default=".")
    new_parser.set_defaults(func=clear)

    parsed_args = parser.parse_args(args)
    try:
        # Launch the function corresponding to the given subcommand.
        kwargs = vars(parsed_args)
        func = kwargs.pop("func")
    except KeyError:
        # No subcommand passed.
        parser.print_help()
        return
    func(**kwargs)


def new(path: Path) -> None:
    """Implement `mcq new` command."""
    template = Path(__file__).resolve().parent / "template"
    if path.exists():
        print(f"ERROR: path {path} already exists.", file=sys.stderr)
        sys.exit(1)
    else:
        shutil.copytree(template, path)
        print(f"Success: a new MCQ was created at {path}.")


def clear(path: Path) -> None:
    """Implement `mcq clear` command."""
    try:
        ptyxfile_path = get_ptyxfile_path(path)
    except FileNotFoundError:
        print("\33[31m[Error]\33[0m No ptyx file found.")
        return
    filename = ptyxfile_path.name
    for directory in (".scan", ".compile"):
        try:
            shutil.rmtree(directory)
        except FileNotFoundError:
            print(f"Info: '{directory}' not found...")
    for filepath in (
        ptyxfile_path.with_suffix(".pdf"),
        ptyxfile_path.with_suffix(".ptyx.autoqcm.config.json"),
        ptyxfile_path.with_name(f".{filename}.plain-ptyx"),
        ptyxfile_path.with_name(f"{ptyxfile_path.stem}-corr.pdf"),
    ):
        try:
            unlink(filepath)
        except FileNotFoundError:
            print(f"Info: '{filepath}' not found...")
    print("\33[32m[OK]\33[0m Directory cleared.")


if __name__ == "__main__":
    main()



import subprocess
import tempfile
from os.path import join

from PIL import Image
from numpy import ndarray, int8

from ptyx_mcq.scan.color import Color, RGB

FloatPosition = tuple[float, float]


# TODO: rewrite this using a class.


# noinspection PyDefaultArgument
def color2debug(
    array: ndarray = None,
    from_: FloatPosition = None,
    to_: FloatPosition = None,
    color: RGB = Color.red,
    display: bool = True,
    thickness: int = 2,
    fill=False,
    _d={},
    wait=True,
):
    """Display picture with a red (by default) rectangle for debugging.
    The annotated picture can be displayed immediately (default), or the annotation can be memorized,
    if `display` is set to `False`.
    It allows to display several annotations at once.

    Launch this function without argument to clear all previous stored annotations.

    Parameters:
    `array` is an array containing the image data (image must be gray mode,
    each pixel represented by a float from 0 (black) to 1 (white)).
    `from_` represent one corner of the red rectangle.
    `to_` represent opposite corner of the red rectangle.
    `color` is given as a RGB tuple ([0-255], [0-255], [0-255]).
    `fill` (True|False) indicates if the rectangle should be filled.

    Usage: color2debug((0,0), (200,10), color=(255, 0, 255))

    If you need to display `n` rectangles, call `color2debug()` with
    `display=False` for the first `n-1` rectangles, and then with
    `display=True` for the last rectangle.

    `_d` is used internally to store values between two runs, if display=False.

    NOTA:
    - `feh` must be installed.
      On Ubuntu/Debian: sudo apt-get install feh
    - Left-dragging the picture with mouse inside feh removes blurring/anti-aliasing,
      making visual debugging a lot easier.
    """
    if array is None:
        _d.clear()
        return
    ID = id(array)
    if ID not in _d:
        # Load image only if not loaded previously.
        # .astype(int8) will make more arrays representable.
        _d[ID] = Image.fromarray((255 * array).astype(int8)).convert("RGB")
    rgb = _d[ID]
    height, width = array.shape
    if from_ is not None:
        if to_ is None:
            to_ = from_
        i1, j1 = from_
        i2, j2 = to_
        if i2 is None:
            i2 = height - 1
        if j2 is None:
            j2 = width - 1
        pix = rgb.load()
        imin, imax = int(min(i1, i2)), int(max(i1, i2))
        jmin, jmax = int(min(j1, j2)), int(max(j1, j2))

        def set_pix(i: int, j: int, color: RGB) -> None:
            """Set safely pixel color (if `i` or `j` is incorrect, do nothing)."""
            if 0 <= i < height and 0 <= j < width:
                pix[j, i] = color

        if fill:
            for i in range(imin, imax + 1):
                for j in range(jmin, jmax + 1):
                    set_pix(i, j, color)
        else:
            # left and right sides of rectangle
            for i in range(imin, imax + 1):
                for j in range(jmin, jmin + thickness):
                    set_pix(i, j, color)
                for j in range(jmax + 1 - thickness, jmax + 1):
                    set_pix(i, j, color)
            # top and bottom sides of rectangle
            for j in range(jmin, jmax + 1):
                for i in range(imin, imin + thickness):
                    set_pix(i, j, color)
                for i in range(imax + 1 - thickness, imax + 1):
                    set_pix(i, j, color)

    if display:
        del _d[ID]
        if subprocess.call(["which", "feh"]) != 0:
            raise RuntimeError(
                "The `feh` command is not found, please " "install it (`sudo apt install feh` on Ubuntu)."
            )
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = join(tmpdirname, "test.png")
            rgb.save(path)
            process: subprocess.CompletedProcess | subprocess.Popen
            if wait:
                process = subprocess.run(["feh", "-F", path])
            else:
                process = subprocess.Popen(["feh", "-F", path], stdin=subprocess.DEVNULL)
            input("-- pause --\n")
            return process



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 23 22:54:07 2020

@author: nicolas
"""

import subprocess
from pathlib import Path
from shutil import rmtree
from os import listdir, mkdir, rename
from os.path import join, basename


PIC_EXTS = (".jpg", ".jpeg", ".png")


def run(cmd):
    """Run command as a subprocess, raising an Python error if it fails."""
    return subprocess.run(cmd, check=True, stdout=subprocess.PIPE)


def _extract_pictures(pdf_path, dest, page=None):
    """Extract all pictures from pdf file in given `dest` directory."""
    # pdfimages `-all` : keep image native format (for jpg, png ans some other formats).
    cmd = ["pdfimages", "-all", pdf_path, join(dest, "pic")]
    if page is not None:
        p = str(page)
        cmd = cmd[:1] + ["-f", p, "-l", p] + cmd[1:]
    # ~ print(cmd)
    run(cmd)


def _export_pdf_to_jpg(pdf_path, dest, page=None):
    print("Convert PDF to JPG, please wait...")
    cmd = [
        "gs",
        "-dNOPAUSE",
        "-dBATCH",
        "-sDEVICE=jpeg",
        "-r200",
        "-sOutputFile=" + join(dest, "%03d.jpg"),
        pdf_path,
    ]
    if page is not None:
        cmd = cmd[:1] + ["-dFirstPage=%s" % page, "-dLastPage=%s" % page] + cmd[1:]
    run(cmd)


def extract_pdf_pictures(pdf_file: Path, dest: Path, page: int = None):
    """Clear `dest` folder, then extract all pages of the pdf files inside."""
    rmtree(dest, ignore_errors=True)
    mkdir(dest)
    tmp_dir = join(dest, ".tmp")
    print(f"Extracting all images from {basename(pdf_file)!r}, please wait...")
    rmtree(tmp_dir, ignore_errors=True)
    mkdir(tmp_dir)
    _extract_pictures(pdf_file, tmp_dir, page)
    # PDF may contain special files (OCR...) we can't handle.
    # In that case, we will rasterize pdf, using Ghostscript.
    pics = listdir(tmp_dir)
    if not all(any(f.endswith(ext) for ext in PIC_EXTS) for f in pics):
        rmtree(tmp_dir)
        mkdir(tmp_dir)
        _export_pdf_to_jpg(pdf_file, tmp_dir, page)
        pics = listdir(tmp_dir)
    for pic in pics:
        rename(join(tmp_dir, pic), join(dest, pic))
    rmtree(tmp_dir)


# def pdf2pic(*pdf_files: str, dest: str, page=None):
#    "Clear `dest` folder, then extract all pages of the pdf files inside."
#    rmtree(dest)
#    mkdir(dest)
#    tmp_dir = join(dest, '.tmp')
#    for pdf in pdf_files:
#        print(f'Extracting all images from {basename(pdf)!r}, please wait...')
#        rmtree(tmp_dir, ignore_errors=True)
#        mkdir(tmp_dir)
#        _extract_pictures(pdf, tmp_dir, page)
#        # PDF may contain special files (OCR...) we can't handle.
#        # In that case, we will rasterize pdf, using Ghostscript.
#        pics = listdir(tmp_dir)
#        if not all(any(f.endswith(ext) for ext in PIC_EXTS) for f in pics):
#            rmtree(tmp_dir)
#            mkdir(tmp_dir)
#            _export_pdf_to_jpg(pdf, tmp_dir, page)
#            pics = listdir(tmp_dir)
#        for pic in pics:
#            rename(join(tmp_dir, pic), join(dest, f'f-{pdf}-{pic}'))
#    rmtree(tmp_dir)
#
# def extract_pictures_from_pdf(source: str, dest: str):
#    "Extract in `dest` directory all pictures from the pdf files found in the" \
#    "`source` directory."
#    # If images are already cached in `.scan` directory, this step will be skipped.
#    pdf_files = glob(join(source, '**/*.pdf'), recursive=True)
#    # ~ pdf_files = [join(INPUT_DIR, name) for name in listdir(INPUT_DIR) if name.endswith('.pdf')]
#    total_page_number = sum(number_of_pages(pdf) for pdf in pdf_files)
#
#    if len(listdir(dest)) != total_page_number:
#        pdf2pic(*pdf_files, dest=dest)
#    else:
#        print("Info: No new pdf file detected.")


def number_of_pages(pdf_path: Path) -> int:
    """Return the number of pages of the pdf."""
    cmd = ["pdfinfo", pdf_path]
    # An example of pdfinfo output:
    # ...
    # JavaScript:     no
    # Pages:          19
    # Encrypted:      no
    # ...
    output = run(cmd).stdout.decode("utf-8").split()
    return int(output[output.index("Pages:") + 1])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ptyx_mcq', 'ptyx_mcq.compile', 'ptyx_mcq.scan', 'ptyx_mcq.tools']

package_data = \
{'': ['*'],
 'ptyx_mcq': ['bin/*',
              'doc/*',
              'template/*',
              'template/questions/*',
              'template/scan/*']}

install_requires = \
['Pillow>=9.2,<10.0',
 'numpy>=1.23.0,<2.0.0',
 'pdf2image>=1.16.0,<2.0.0',
 'ptyx>=22.6,<23.0',
 'sympy>=1.10.1,<2.0.0',
 'types-Pillow>=9.0.20,<10.0.0']

entry_points = \
{'console_scripts': ['mcq = ptyx_mcq.cli:main'],
 'ptyx.extensions': ['mcq = ptyx_mcq']}

setup_kwargs = {
    'name': 'ptyx-mcq',
    'version': '22.4',
    'description': 'pTyX is a python precompiler for LaTeX.',
    'long_description': "pTyX MCQ Extension\n==================\n\n***Disclaimer:*** installation is broken for now (work in progress...)\n\nIf you're interested, **please wait a few months** before it becomes fully usable.\n\n(For now, the following document is only a draft documenting future usage.)\n\nOverview\n--------\npTyX is a LaTeX precompiler, written in Python.\npTyX enables to generate LaTeX documents, using custom commands or plain python code.\nOne single pTyX file may generate many latex documents, with different values.\nI developed and used pTyX to make several versions of a same test in exams,\nfor my student, to discourage cheating.\nSince it uses sympy library, pTyX has symbolic calculus abilities too.\n\nThe `pTyX MCQ extension` makes it easy to use pTyX to generate Multiple Choice Questions as pdf documents.\nThe students MCQ can then be scanned and automatically corrected and scored.\n\nInstallation\n------------\n\nObviously, pTyX needs a working Python installation.\nPython version 3.8 (at least) is required for pTyX MCQ to run.\n\nCurrently, pTyX is only supported on GNU/Linux.\n\nThe easiest way to install it is to use pip.\n\n    $ pip install --user ptyx_mcq\n\nUsage\n-----\n\nTo generate a template, run:\n\n    $ mcq new new_folder\n\nThis will generate a `new_folder/new.ptyx` file, which is the main configuration file.\n\nThis will also create a `new_folder/questions/` folder, where you should put all the exercises, \nas `.txt` files. \n\nSee the next section (*MCQ file format*) for more information about those files format.\n\nTo compile the template, run:\n\n    $ mcq make\n\nFor more options:\n\n    $ mcq make --help\n\nTo automatically corrected the scanned students MCQs, but them as a pdf inside `new_folder/scan`.\n\nThen run:\n    \n    $ mcq scan\n\n\nMCQ file format\n---------------\n\nWhen running `mcq new`, a template folder will be generated, including a `new.ptyx` file.\n\n(More to come...)\n",
    'author': 'Nicolas Pourcelot',
    'author_email': 'nicolas.pourcelot@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/wxgeo/ptyx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 27 22:40:40 2020

@author: nicolas
"""
import builtins
from os.path import expanduser
from pathlib import Path
import pickle


def search_by_extension(directory: Path, ext: str) -> Path:
    """Search for a file with extension `ext` in given directory.

    Search is NOT case sensible.
    If no or more than one file is found, an error is raised.
    """
    ext = ext.lower()
    paths = [pth for pth in directory.glob("*") if pth.name.lower().endswith(ext)]
    if not paths:
        raise FileNotFoundError(f"No {ext!r} file found in that directory: {directory} ! ")
    elif len(paths) > 1:
        raise RuntimeError(
            f"Several {ext!r} file found in that directory: {directory} ! "
            "Keep only one of them and delete all the others (or rename their extensions)."
        )
    return paths[0]


def print_framed_msg(msg: str) -> None:
    decoration = max(len(line) for line in msg.split("\n")) * "-"
    print(decoration)
    print(msg)
    print(decoration)


def tmp_store(obj):
    """For debugging."""
    with open(expanduser("~/tmp.pickle"), "wb") as f:
        pickle.dump(obj, f)


def tmp_load():
    """For debugging."""
    with open(expanduser("~/tmp.pickle"), "rb") as f:
        return pickle.load(f)


def round(f, n=None):
    # PEP3141 compatible round() implementation.
    # round(f) should return an integer, but the problem is
    # __builtin__.round(f) doesn't return an int if type(f) is np.float64.
    # See: https://github.com/numpy/numpy/issues/11810
    return int(builtins.round(f)) if n is None else builtins.round(f, n)


def levenshtein_distance(string1: str, string2: str) -> int:
    """Return the Levenshtein distance between the two strings.

    The Levenshtein distance between two words is the minimum number
    of single-character edits (insertions, deletions or substitutions)
    required to change one word into the other.
    """
    # cf. https://en.wikipedia.org/wiki/Levenshtein_distance
    # Create two work vectors of integer distances
    # Initialize previous_row (the previous row of distances)
    # this row is A[0][i]: edit distance from an empty s to t;
    # that distance is the number of characters to append to  s to make t.
    m = len(string1)
    n = len(string2)

    previous_row = list(range(n + 1))
    current_row = [0 for _ in previous_row]

    for i in range(m):
        # Calculate current_row (current row distances) from the previous row previous_row
        # first element of current_row is A[i + 1][0]
        # edit distance is delete (i + 1) chars from s to match empty t
        current_row[0] = i + 1
        # use formula to fill in the rest of the row
        for j in range(n):
            # calculating costs for A[i + 1][j + 1]
            deletionCost = previous_row[j + 1] + 1
            insertionCost = current_row[j] + 1
            if string1[i] == string2[j]:
                substitutionCost = previous_row[j]
            else:
                substitutionCost = previous_row[j] + 1
            current_row[j + 1] = min(deletionCost, insertionCost, substitutionCost)
        # copy current_row to previous_row for next iteration
        # since data in current_row is always invalidated, a swap without copy is more efficient
        previous_row, current_row = current_row, previous_row
    # after the last swap, the results of current_row are now in previous_row
    return previous_row[n]
